"""Icon accessor with auto-theming."""

import enum
import functools
from typing import assert_never

from PySide6.QtGui import QIcon

from usdb_syncer import settings


class Icon(enum.Enum):
    """Available icons with auto-theming."""

    ABORT = "minus-circle.png", "minus-circle-white.svg"
    ARC = "arc.png", None
    ARTIST = "artist.png", "artist-white.svg"
    AUDIO = "audio.png", "music-white.svg"
    AUDIO_SAMPLE = "sample.png", "music-note-white.svg"
    BACKGROUND = "background.png", "image-area-white.svg"
    BRAVE = "brave.png", None
    BUG = "bug.png", "bug-white.svg"
    CALENDAR = "calendar.png", "calendar-white.svg"
    CHECK = "tick.png", "check-white.svg"
    CHECK_FOR_UPDATE = "check_for_update.png", "sync-white.svg"
    CHEVRON_DOWN = None, "chevron-down-white.svg"
    CHEVRON_LEFT = None, "chevron-left-white.svg"
    CHEVRON_RIGHT = None, "chevron-right-white.svg"
    CHEVRON_UP = None, "chevron-up-white.svg"
    CHROME = "chrome.png", None
    CHROMIUM = "chromium.png", None
    COMMENT = "balloon.png", "comment-white.svg"
    COVER = "cover.png", "image-white.svg"
    CREATOR = "quill.png", "feather-white.svg"
    CUSTOM_DATA = "drawer.png", "dresser-white.svg"
    DATABASE = "database.png", "database-white.svg"
    DELETE = "bin.png", "delete-white.svg"
    DOWNLOAD = "status.png", "download-white.svg"
    EDGE = "edge.png", None
    EDITION = "edition.png", "edition-white.svg"
    ERROR = "error.png", "close-octagon-white.svg"
    FILE_EXPORT = "document-export.png", "file-export-white.svg"
    FILE_IMPORT = "document-import.png", "file-import-white.svg"
    FIREFOX = "firefox.png", None
    GENRE = "spectrum-absorption.png", "cards-white.svg"
    GOLDEN_NOTES = "golden_notes.png", "gold-white.svg"
    ID = "id.png", "database-marker-white.svg"
    INFO = "info.png", "information-white.svg"
    KAREDI = "karedi.png", None
    LANGUAGE = "language.png", "translate-variant-white.svg"
    LIBREWOLF = "librewolf.png", None
    LOG = "log.png", "receipt-text-white.svg"
    MENU_DOWN = None, "menu-down-white.svg"
    MENU_LEFT = None, "menu-left-white.svg"
    MENU_RIGHT = None, "menu-right-white.svg"
    MENU_UP = None, "menu-up-white.svg"
    META_TAGS = "tag-hash.png", "pound-box-white.svg"
    OCTO_BROWSER = "octo_browser.png", None
    OPEN_SONG_WITH = "music--arrow.png", "folder-play-white.svg"
    OPERA = "opera.png", None
    OPERA_GX = "opera_gx.png", None
    PAUSE_LOCAL = "control-pause-local.png", "pause-white.svg"
    PAUSE_REMOTE = "control-pause.png", "pause-white.svg"
    PERFORMOUS = "performous.png", None
    PIN = "pin.png", "pin-white.svg"
    PLAY_LOCAL = "control-play-local.png", "play-white.svg"
    PLAY_REMOTE = "control-play.png", "play-network-white.svg"
    RADIOBOX_BLANK = None, "radiobox-blank-white.svg"
    RADIOBOX_MARKED = None, "radiobox-marked-white.svg"
    RATING = "rating.png", "podium-white.svg"
    REPORT = "report.png", "notebook-white.svg"
    SAFARI = "safari.png", None
    SAVED_SEARCH = "heart.png", "heart-white.svg"
    SETTINGS = "cog.png", "cog-white.svg"
    SONG_FOLDER = "folder_note.png", "folder-music-white.svg"
    TAGS = "price-tag.png", "tag-white.svg"
    TEXT = "text.png", "file-document-white.svg"
    TITLE = "title.png", "format-title-white.svg"
    ULTRASTAR_MANAGER = "ultrastar-manager.png", None
    USDB = "faviconUSDB.png", "usdb-white.svg"
    USDX = "usdx.png", None
    VIDEO = "video.png", "filmstrip-white.svg"
    VIEWS = "views.png", "views-white.svg"
    VIVALDI = "vivaldi.png", None
    VOCALUXE = "vocaluxe.png", None
    WARNING = "warning.png", "alert-white.svg"
    YASS = "yass.png", None
    YASS_RELOADED = "yass-reloaded.png", None

    def __init__(self, colored_name: str | None, white_name: str | None):
        colored = colored_name or white_name
        white = white_name or colored_name
        assert colored is not None and white is not None, (
            f"{self.__class__.__name__}.{self.name}: At least one value must be "
            "supplied"
        )
        self.colored_name = colored
        self.white_name = white

    def icon(self, theme: settings.Theme | None = None) -> QIcon:
        colored = (theme or settings.get_theme()) == settings.Theme.SYSTEM
        return _icon(self.colored_name if colored else self.white_name)


def browser_icon(browser: settings.Browser) -> QIcon | None:  # noqa: C901
    match browser:
        case settings.Browser.NONE:
            return None
        case settings.Browser.ARC:
            icon = Icon.ARC
        case settings.Browser.BRAVE:
            icon = Icon.BRAVE
        case settings.Browser.CHROME:
            icon = Icon.CHROME
        case settings.Browser.CHROMIUM:
            icon = Icon.CHROMIUM
        case settings.Browser.EDGE:
            icon = Icon.EDGE
        case settings.Browser.FIREFOX:
            icon = Icon.FIREFOX
        case settings.Browser.LIBREWOLF:
            icon = Icon.LIBREWOLF
        case settings.Browser.OCTO_BROWSER:
            icon = Icon.OCTO_BROWSER
        case settings.Browser.OPERA:
            icon = Icon.OPERA
        case settings.Browser.OPERA_GX:
            icon = Icon.OPERA_GX
        case settings.Browser.SAFARI:
            icon = Icon.SAFARI
        case settings.Browser.VIVALDI:
            icon = Icon.VIVALDI
        case _ as unreachable:
            assert_never(unreachable)
    return icon.icon()


@functools.cache
def _icon(name: str) -> QIcon:
    return QIcon(f":/icons/{name}")



"""Database utilities."""

from __future__ import annotations

import contextlib
import enum
import json
import sqlite3
import threading
import time
from collections import defaultdict
from collections.abc import Generator, Iterable, Iterator
from importlib import resources
from pathlib import Path
from typing import Any, ClassVar, assert_never, cast

import attrs
from more_itertools import batched

from usdb_syncer import SongId, SyncMetaId, errors
from usdb_syncer.logger import logger

from . import sql

SCHEMA_VERSION = 6

# https://www.sqlite.org/limits.html
_SQL_VARIABLES_LIMIT = 32766


class _SqlCache:
    _cache: ClassVar[dict[str, str]] = {}

    @classmethod
    def get(cls, name: str, cache: bool = True) -> str:
        if (stmt := cls._cache.get(name)) is None:
            stmt = resources.files(sql).joinpath(name).read_text("utf8")
            if cache:
                cls._cache[name] = stmt
        return stmt


class _LocalConnection(threading.local):
    """A thread-local database connection."""

    connection: sqlite3.Connection | None = None


class _DbState:
    """Singleton for managing the global database connection."""

    _local: _LocalConnection = _LocalConnection()
    trace_sql: bool = False

    @classmethod
    def connect(cls, db_path: Path | str) -> None:
        if cls._local.connection:
            raise errors.AlreadyConnectedError()
        cls._local.connection = sqlite3.connect(
            db_path, check_same_thread=False, isolation_level=None, timeout=20
        )
        thread = threading.current_thread().name
        logger.debug(f"Connected to database at '{db_path}' on thread {thread}.")
        if cls.trace_sql:
            cls._local.connection.set_trace_callback(logger.debug)
        _validate_schema(cls._local.connection)

    @classmethod
    def connection(cls) -> sqlite3.Connection:
        if cls._local.connection is None:
            raise errors.NotConnectedError()
        return cls._local.connection

    @classmethod
    def close(cls) -> None:
        if _DbState._local.connection is not None:
            _DbState._local.connection.close()
            _DbState._local.connection = None
            thread = threading.current_thread().name
            logger.debug(f"Closed database connection on thread {thread}.")


@contextlib.contextmanager
def transaction() -> Generator[None, None, None]:
    try:
        _DbState.connection().execute("BEGIN IMMEDIATE")
        yield None
    except Exception:
        _DbState.connection().rollback()
        raise
    _DbState.connection().commit()


def _validate_schema(connection: sqlite3.Connection) -> None:
    meta_table = connection.execute(
        "SELECT 1 FROM sqlite_master WHERE type = 'table' AND name = 'meta'"
    ).fetchone()
    if meta_table is None:
        version = 0
    else:
        row = connection.execute("SELECT version FROM meta WHERE id = 1").fetchone()
        if not row or row[0] > SCHEMA_VERSION:
            raise errors.UnknownSchemaError
        version = row[0]
    for ver in range(version + 1, SCHEMA_VERSION + 1):
        connection.executescript(_SqlCache.get(f"{ver}_migration.sql", cache=False))
        logger.debug(f"Database migrated to version {ver}.")
    if version < SCHEMA_VERSION:
        connection.execute(
            "INSERT INTO meta (id, version, ctime) VALUES (1, :version, :ctime) "
            "ON CONFLICT (id) DO UPDATE SET version = :version",
            {"version": SCHEMA_VERSION, "ctime": int(time.time() * 1_000_000)},
        )
    connection.executescript(_SqlCache.get("setup_session_script.sql", cache=False))


def connect(db_path: Path | str) -> None:
    _DbState.connect(db_path)


def close() -> None:
    _DbState.close()


@contextlib.contextmanager
def managed_connection(db_path: Path | str) -> Generator[None, None, None]:
    try:
        _DbState.connect(db_path)
        yield None
    finally:
        _DbState.close()


def set_trace_sql(trace_sql: bool) -> None:
    """Set SQL logging for future connections."""
    _DbState.trace_sql = trace_sql


class DownloadStatus(enum.IntEnum):
    """Status of song in download queue."""

    NONE = 0
    PENDING = enum.auto()
    DOWNLOADING = enum.auto()
    FAILED = enum.auto()

    def __str__(self) -> str:
        match self:
            case DownloadStatus.NONE:
                return ""
            case DownloadStatus.PENDING:
                return "Pending"
            case DownloadStatus.DOWNLOADING:
                return "Downloading"
            case DownloadStatus.FAILED:
                return "Failed"
            case _ as unreachable:
                assert_never(unreachable)

    def can_be_downloaded(self) -> bool:
        return self in (DownloadStatus.NONE, DownloadStatus.FAILED)

    def can_be_aborted(self) -> bool:
        return self in (DownloadStatus.PENDING, DownloadStatus.DOWNLOADING)


class SongOrder(enum.Enum):
    """Attributes songs can be sorted by."""

    NONE = 0
    SAMPLE_URL = enum.auto()
    SONG_ID = enum.auto()
    ARTIST = enum.auto()
    TITLE = enum.auto()
    EDITION = enum.auto()
    LANGUAGE = enum.auto()
    GOLDEN_NOTES = enum.auto()
    RATING = enum.auto()
    VIEWS = enum.auto()
    YEAR = enum.auto()
    GENRE = enum.auto()
    CREATOR = enum.auto()
    TAGS = enum.auto()
    PINNED = enum.auto()
    TXT = enum.auto()
    AUDIO = enum.auto()
    VIDEO = enum.auto()
    COVER = enum.auto()
    BACKGROUND = enum.auto()
    STATUS = enum.auto()

    def sql(self) -> str | None:  # noqa: C901
        match self:
            case SongOrder.NONE:
                return None
            case SongOrder.SAMPLE_URL:
                return (
                    "CASE WHEN session_usdb_song.is_playing = true THEN 0"
                    " WHEN audio.sync_meta_id IS NOT NULL THEN 1"
                    " WHEN usdb_song.sample_url != '' THEN 2 ELSE 3 END"
                )
            case SongOrder.SONG_ID:
                return "usdb_song.song_id"
            case SongOrder.ARTIST:
                return "usdb_song.artist"
            case SongOrder.TITLE:
                return "usdb_song.title"
            case SongOrder.EDITION:
                return "usdb_song.edition"
            case SongOrder.LANGUAGE:
                return "usdb_song.language"
            case SongOrder.GOLDEN_NOTES:
                return "usdb_song.golden_notes"
            case SongOrder.RATING:
                return "usdb_song.rating"
            case SongOrder.VIEWS:
                return "usdb_song.views"
            case SongOrder.YEAR:
                return "usdb_song.year"
            case SongOrder.GENRE:
                return "usdb_song.genre"
            case SongOrder.CREATOR:
                return "usdb_song.creator"
            case SongOrder.TAGS:
                return "usdb_song.tags"
            case SongOrder.PINNED:
                return "sync_meta.pinned"
            case SongOrder.TXT:
                return "txt.sync_meta_id IS NULL"
            case SongOrder.AUDIO:
                return "audio.sync_meta_id IS NULL"
            case SongOrder.VIDEO:
                return "video.sync_meta_id IS NULL"
            case SongOrder.COVER:
                return "cover.sync_meta_id IS NULL"
            case SongOrder.BACKGROUND:
                return "background.sync_meta_id IS NULL"
            case SongOrder.STATUS:
                return (
                    "coalesce(session_usdb_song.status, sync_meta.mtime,"
                    # max integer in SQLite
                    " 9223372036854775807)"
                )


@attrs.define
class SearchBuilder:
    """Helper for building a where clause to find songs."""

    order: SongOrder = SongOrder.NONE
    descending: bool = False
    text: str = ""
    artists: list[str] = attrs.field(factory=list)
    titles: list[str] = attrs.field(factory=list)
    editions: list[str] = attrs.field(factory=list)
    ratings: list[int] = attrs.field(factory=list)
    statuses: list[DownloadStatus] = attrs.field(factory=list)
    languages: list[str] = attrs.field(factory=list)
    views: list[tuple[int, int | None]] = attrs.field(factory=list)
    years: list[int] = attrs.field(factory=list)
    genres: list[str] = attrs.field(factory=list)
    creators: list[str] = attrs.field(factory=list)
    golden_notes: bool | None = None
    downloaded: bool | None = None

    def filters(self) -> Iterator[str]:
        if _fts5_phrases(self.text):
            yield (
                "usdb_song.song_id IN (SELECT rowid FROM fts_usdb_song WHERE"
                " fts_usdb_song MATCH ?)"
            )
        for vals, col in (
            (self.artists, "usdb_song.artist"),
            (self.titles, "usdb_song.title"),
            (self.editions, "usdb_song.edition"),
            (self.ratings, "usdb_song.rating"),
            (self.years, "usdb_song.year"),
            (self.statuses, "session_usdb_song.status"),
        ):
            if vals:
                yield _in_values_clause(col, cast(list, vals))
        if self.languages:
            yield (
                "usdb_song.song_id IN (SELECT song_id FROM usdb_song_language WHERE"
                f" {_in_values_clause('language', self.languages)})"
            )
        if self.genres:
            yield (
                "usdb_song.song_id IN (SELECT song_id FROM usdb_song_genre WHERE"
                f" {_in_values_clause('genre', self.genres)})"
            )
        if self.creators:
            yield (
                "usdb_song.song_id IN (SELECT song_id FROM usdb_song_creator WHERE"
                f" {_in_values_clause('creator', self.creators)})"
            )
        if self.views:
            yield _in_ranges_clause("usdb_song.views", self.views)
        if self.golden_notes is not None:
            yield "usdb_song.golden_notes = ?"
        if self.downloaded is not None:
            yield f"sync_meta.sync_meta_id IS {'NOT ' if self.downloaded else ''}NULL"

    def _where_clause(self) -> str:
        where = " AND ".join(self.filters())
        return f" WHERE {where}" if where else ""

    def _order_by_clause(self) -> str:
        if sql := self.order.sql():
            return f" ORDER BY {sql} {'DESC' if self.descending else 'ASC'}"
        return ""

    def parameters(self) -> Iterator[str | int | bool]:
        if text := _fts5_phrases(self.text):
            yield text
        yield from self.artists
        yield from self.titles
        yield from self.editions
        yield from self.ratings
        yield from self.years
        yield from self.statuses
        yield from self.languages
        yield from self.genres
        yield from self.creators
        for min_views, max_views in self.views:
            yield min_views
            if max_views is not None:
                yield max_views
        if self.golden_notes is not None:
            yield self.golden_notes

    def statement(self) -> str:
        select_from = _SqlCache.get("select_song_id.sql")
        where = self._where_clause()
        order_by = self._order_by_clause()
        return f"{select_from}{where}{order_by}"

    def to_json(self) -> str:
        return json.dumps(self, cls=_SearchEnoder)

    @classmethod
    def from_json(cls, json_str: str) -> SearchBuilder | None:
        fields = attrs.fields(cls)
        try:
            dct = json.loads(json_str)
            dct[fields.order.name] = SongOrder(dct[fields.order.name])
            dct[fields.statuses.name] = [
                DownloadStatus(s) for s in dct[fields.statuses.name]
            ]
            dct[fields.views.name] = [tuple(item) for item in dct[fields.views.name]]
            return cls(**dct)
        except (
            json.decoder.JSONDecodeError,
            UnicodeDecodeError,
            TypeError,
            KeyError,
            ValueError,
        ):
            logger.exception(None)
        return None


class _SearchEnoder(json.JSONEncoder):
    """Custom encoder for a search."""

    def default(self, o: Any) -> Any:
        if isinstance(o, SearchBuilder):
            return attrs.asdict(o, recurse=False)
        if isinstance(o, enum.Enum):
            return o.value
        return super().default(o)


@attrs.define
class SavedSearch:
    """A search saved to the database by the user."""

    name: str
    search: SearchBuilder
    is_default: bool = False
    subscribed: bool = False

    def insert(self) -> None:
        self.name = (
            _DbState.connection()
            .execute(
                _SqlCache.get("insert_saved_search.sql"),
                {
                    "name": self.name,
                    "search": self.search.to_json(),
                    "is_default": self.is_default,
                    "subscribed": self.subscribed,
                },
            )
            .fetchone()
        )[0]

    def delete(self) -> None:
        _DbState.connection().execute(
            "DELETE FROM saved_search WHERE name = ?", (self.name,)
        )

    @classmethod
    def get(cls, name: str) -> SavedSearch | None:
        stmt = (
            "SELECT name, search, is_default, subscribed FROM saved_search"
            " WHERE name = ?"
        )
        row = _DbState.connection().execute(stmt, (name,)).fetchone()
        if row and (search := cls._validate_saved_search_row(*row)):
            return search
        return None

    @classmethod
    def get_default(cls) -> SavedSearch | None:
        stmt = (
            "SELECT name, search, is_default, subscribed FROM saved_search"
            " WHERE is_default = true"
        )
        row = _DbState.connection().execute(stmt).fetchone()
        if row and (search := cls._validate_saved_search_row(*row)):
            return search
        return None

    def update(self, new_name: str | None = None) -> None:
        self.name = (
            _DbState.connection()
            .execute(
                _SqlCache.get("update_saved_search.sql"),
                {
                    "old_name": self.name,
                    "new_name": new_name or self.name,
                    "search": self.search.to_json(),
                    "is_default": self.is_default,
                    "subscribed": self.subscribed,
                },
            )
            .fetchone()
        )[0]

    @classmethod
    def load_saved_searches(
        cls, subscribed_only: bool = False
    ) -> Iterable[SavedSearch]:
        stmt = (
            "SELECT name, search, is_default, subscribed FROM saved_search"
            f"{' WHERE subscribed' if subscribed_only else ''} ORDER BY name"
        )
        return (
            search
            for row in _DbState.connection().execute(stmt).fetchall()
            if (search := cls._validate_saved_search_row(*row))
        )

    @classmethod
    def _validate_saved_search_row(
        cls, name: str, json_str: str, is_default: int, subscribed: int
    ) -> SavedSearch | None:
        if search := SearchBuilder.from_json(json_str):
            return cls(name, search, bool(is_default), bool(subscribed))
        _DbState.connection().execute(
            "DELETE FROM saved_search WHERE name = ?", (name,)
        )
        logger.warning(f"Dropped invalid saved search '{name}'.")
        return None

    @classmethod
    def get_subscribed_song_ids(cls) -> Iterable[SongId]:
        if not (searches := list(cls.load_saved_searches(subscribed_only=True))):
            return []
        for search in searches:
            search.search.order = SongOrder.NONE
        stmt = "\nUNION\n".join(s.search.statement() for s in searches)
        params = tuple(p for s in searches for p in s.search.parameters())
        return (SongId(r[0]) for r in _DbState.connection().execute(stmt, params))


def _in_values_clause(attribute: str, values: list) -> str:
    return f"{attribute} IN ({', '.join('?' * len(values))})"


def _in_ranges_clause(attribute: str, values: list[tuple[int, int | None]]) -> str:
    return " OR ".join(
        f"{attribute} >= ?{'' if val[1] is None else f' AND {attribute} < ?'}"
        for val in values
    )


def _fts5_phrases(text: str) -> str:
    """Turns each whitespace-separated word into an FTS5 prefix phrase."""
    return " ".join(f'"{s}"*' for s in text.replace('"', "").split(" ") if s)


def _fts5_start_phrase(text: str) -> str:
    """Turns the entire string into an FTS5 initial phrase."""
    return f'''^ "{text.replace('"', "")}"'''


# UsdbSong


def get_usdb_song(song_id: SongId) -> tuple | None:
    stmt = f"{_SqlCache.get('select_usdb_song.sql')} WHERE usdb_song.song_id = ?"
    return _DbState.connection().execute(stmt, (song_id,)).fetchone()


def delete_usdb_song(song_id: SongId) -> None:
    _DbState.connection().execute("DELETE FROM usdb_song WHERE song_id = ?", (song_id,))


@attrs.define(frozen=True, slots=False)
class UsdbSongParams:
    """Parameters for inserting or updating a USDB song."""

    song_id: SongId
    artist: str
    title: str
    language: str
    edition: str
    golden_notes: bool
    rating: int
    views: int
    sample_url: str
    year: int | None
    genre: str
    creator: str
    tags: str
    status: DownloadStatus
    is_playing: bool


def upsert_usdb_song(params: UsdbSongParams) -> None:
    stmt = _SqlCache.get("upsert_usdb_song.sql")
    _DbState.connection().execute(stmt, params.__dict__)
    stmt = _SqlCache.get("upsert_session_usdb_song.sql")
    _DbState.connection().execute(stmt, params.__dict__)


def upsert_usdb_songs(params: Iterable[UsdbSongParams]) -> None:
    stmt = _SqlCache.get("upsert_usdb_song.sql")
    _DbState.connection().executemany(stmt, (p.__dict__ for p in params))


def usdb_song_count() -> int:
    return _DbState.connection().execute("SELECT count(*) FROM usdb_song").fetchone()[0]


def max_usdb_song_id() -> SongId:
    row = _DbState.connection().execute("SELECT max(song_id) FROM usdb_song").fetchone()
    return SongId(row[0] or 0)


def delete_all_usdb_songs() -> None:
    _DbState.connection().execute("DELETE FROM usdb_song")


def all_local_usdb_songs() -> Iterable[SongId]:
    stmt = "SELECT DISTINCT song_id FROM sync_meta"
    return (SongId(r[0]) for r in _DbState.connection().execute(stmt))


def all_song_ids() -> Iterable[SongId]:
    rows = _DbState.connection().execute("SELECT song_id FROM usdb_song")
    return (SongId(r[0]) for r in rows)


def search_usdb_songs(search: SearchBuilder) -> Iterable[SongId]:
    rows = _DbState.connection().execute(search.statement(), tuple(search.parameters()))
    return (SongId(r[0]) for r in rows)


def find_similar_usdb_songs(artist: str, title: str) -> Iterable[SongId]:
    stmt = "SELECT rowid FROM fts_usdb_song WHERE artist MATCH ? AND title MATCH ?"
    params = (_fts5_start_phrase(artist), _fts5_start_phrase(title))
    return (SongId(r[0]) for r in _DbState.connection().execute(stmt, params))


# song filters


def upsert_usdb_songs_languages(params: list[tuple[SongId, Iterable[str]]]) -> None:
    _DbState.connection().execute(
        f"DELETE FROM usdb_song_language WHERE {_in_values_clause('song_id', params)}",
        tuple(t[0] for t in params),
    )
    _DbState.connection().executemany(
        "INSERT INTO usdb_song_language (song_id, language) VALUES (?, ?)",
        ((song_id, lang) for song_id, langs in params for lang in langs),
    )


def upsert_usdb_songs_genres(params: list[tuple[SongId, Iterable[str]]]) -> None:
    _DbState.connection().execute(
        f"DELETE FROM usdb_song_genre WHERE {_in_values_clause('song_id', params)}",
        tuple(t[0] for t in params),
    )
    _DbState.connection().executemany(
        "INSERT INTO usdb_song_genre (song_id, genre) VALUES (?, ?)",
        ((song_id, lang) for song_id, langs in params for lang in langs),
    )


def upsert_usdb_songs_creators(params: list[tuple[SongId, Iterable[str]]]) -> None:
    _DbState.connection().execute(
        f"DELETE FROM usdb_song_creator WHERE {_in_values_clause('song_id', params)}",
        tuple(t[0] for t in params),
    )
    _DbState.connection().executemany(
        "INSERT INTO usdb_song_creator (song_id, creator) VALUES (?, ?)",
        ((song_id, lang) for song_id, langs in params for lang in langs),
    )


def usdb_song_artists() -> list[tuple[str, int]]:
    stmt = "SELECT artist, COUNT(*) FROM usdb_song GROUP BY artist ORDER BY artist"
    return _DbState.connection().execute(stmt).fetchall()


def usdb_song_titles() -> list[tuple[str, int]]:
    stmt = "SELECT title, COUNT(*) FROM usdb_song GROUP BY title ORDER BY title"
    return _DbState.connection().execute(stmt).fetchall()


def usdb_song_editions() -> list[tuple[str, int]]:
    stmt = "SELECT edition, COUNT(*) FROM usdb_song GROUP BY edition ORDER BY edition"
    return _DbState.connection().execute(stmt).fetchall()


def usdb_song_languages() -> list[tuple[str, int]]:
    stmt = (
        "SELECT language, COUNT(*) FROM usdb_song_language GROUP BY language ORDER BY"
        " language"
    )
    return _DbState.connection().execute(stmt).fetchall()


def usdb_song_years() -> list[tuple[int, int]]:
    stmt = (
        "SELECT year, COUNT(*) FROM usdb_song WHERE year IS NOT NULL "
        "GROUP BY year ORDER BY year"
    )
    return _DbState.connection().execute(stmt).fetchall()


def usdb_song_genres() -> list[tuple[str, int]]:
    stmt = "SELECT genre, COUNT(*) FROM usdb_song_genre GROUP BY genre ORDER BY genre"
    return _DbState.connection().execute(stmt).fetchall()


def usdb_song_creators() -> list[tuple[str, int]]:
    stmt = (
        "SELECT creator, COUNT(*) FROM usdb_song_creator GROUP BY creator ORDER BY "
        "creator"
    )
    return _DbState.connection().execute(stmt).fetchall()


def search_usdb_song_artists(search: str) -> set[str]:
    stmt = "SELECT artist FROM fts_usdb_song WHERE artist MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


def search_usdb_song_titles(search: str) -> set[str]:
    stmt = "SELECT title FROM fts_usdb_song WHERE title MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


def search_usdb_song_editions(search: str) -> set[str]:
    stmt = "SELECT edition FROM fts_usdb_song WHERE edition MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


def search_usdb_song_languages(search: str) -> set[str]:
    stmt = "SELECT language FROM fts_usdb_song WHERE language MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


def search_usdb_song_years(search: str) -> set[int]:
    stmt = "SELECT year FROM fts_usdb_song WHERE year MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


def search_usdb_song_genres(search: str) -> set[str]:
    stmt = "SELECT genre FROM fts_usdb_song WHERE genre MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


def search_usdb_song_creators(search: str) -> set[str]:
    stmt = "SELECT creator FROM fts_usdb_song WHERE creator MATCH ?"
    rows = _DbState.connection().execute(stmt, (_fts5_phrases(search),)).fetchall()
    return {row[0] for row in rows}


# SyncMeta


def get_in_folder(folder: Path) -> list[tuple]:
    stmt = f"{_SqlCache.get('select_sync_meta.sql')} WHERE path GLOB ? || '/*'"
    return _DbState.connection().execute(stmt, (folder.as_posix(),)).fetchall()


def reset_active_sync_metas(folder: Path) -> None:
    _DbState.connection().execute("DELETE FROM active_sync_meta")
    params = {"folder": folder.as_posix()}
    _DbState.connection().execute(_SqlCache.get("insert_active_sync_metas.sql"), params)


def update_active_sync_metas(folder: Path, song_id: SongId) -> None:
    _DbState.connection().execute(
        "DELETE FROM active_sync_meta WHERE song_id = ?", (song_id,)
    )
    params = {"folder": folder.as_posix(), "song_id": song_id}
    _DbState.connection().execute(_SqlCache.get("insert_active_sync_meta.sql"), params)


@attrs.define(frozen=True, slots=False)
class SyncMetaParams:
    """Parameters for inserting or updating a sync meta."""

    sync_meta_id: SyncMetaId
    song_id: SongId
    path: str
    mtime: int
    meta_tags: str
    pinned: bool


def upsert_sync_meta(params: SyncMetaParams) -> None:
    stmt = _SqlCache.get("upsert_sync_meta.sql")
    _DbState.connection().execute(stmt, params.__dict__)


def upsert_sync_metas(params: Iterable[SyncMetaParams]) -> None:
    stmt = _SqlCache.get("upsert_sync_meta.sql")
    _DbState.connection().executemany(stmt, (p.__dict__ for p in params))


def delete_sync_meta(sync_meta_id: SyncMetaId) -> None:
    _DbState.connection().execute(
        "DELETE FROM sync_meta WHERE sync_meta_id = ?", (sync_meta_id,)
    )


def delete_sync_metas(ids: tuple[SyncMetaId, ...]) -> None:
    for batch in batched(ids, _SQL_VARIABLES_LIMIT):
        id_str = ", ".join("?" for _ in range(len(batch)))
        _DbState.connection().execute(
            f"DELETE FROM sync_meta WHERE sync_meta_id IN ({id_str})", batch
        )


@attrs.define(frozen=True, slots=False)
class CustomMetaDataParams:
    """Parameters for inserting or updating a resource file."""

    sync_meta_id: SyncMetaId
    key: str
    value: str


def upsert_custom_meta_data(params: Iterable[CustomMetaDataParams]) -> None:
    stmt = _SqlCache.get("upsert_custom_meta_data.sql")
    _DbState.connection().executemany(stmt, (p.__dict__ for p in params))


def delete_custom_meta_data(ids: Iterable[SyncMetaId]) -> None:
    for batch in batched(ids, _SQL_VARIABLES_LIMIT):
        id_str = ", ".join("?" for _ in range(len(batch)))
        _DbState.connection().execute(
            f"DELETE FROM custom_meta_data WHERE sync_meta_id IN ({id_str})", batch
        )


def get_custom_data(sync_meta_id: SyncMetaId) -> dict[str, str]:
    return dict(
        _DbState.connection().execute(
            "SELECT key, value FROM custom_meta_data WHERE sync_meta_id = ?",
            (int(sync_meta_id),),
        )
    )


def get_custom_data_map() -> defaultdict[str, set[str]]:
    data: defaultdict[str, set[str]] = defaultdict(set)
    for key, value in _DbState.connection().execute(
        "SELECT DISTINCT key, value FROM custom_meta_data"
    ):
        data[key].add(value)
    return data


# ResourceFile


class ResourceFileKind(str, enum.Enum):
    """Kinds of resource files."""

    TXT = "txt"
    AUDIO = "audio"
    VIDEO = "video"
    COVER = "cover"
    BACKGROUND = "background"


@attrs.define(frozen=True, slots=False)
class ResourceFileParams:
    """Parameters for inserting or updating a resource file."""

    sync_meta_id: SyncMetaId
    kind: ResourceFileKind
    fname: str
    mtime: int
    resource: str


def delete_resource_files(ids: Iterable[tuple[SyncMetaId, ResourceFileKind]]) -> None:
    for batch in batched(ids, _SQL_VARIABLES_LIMIT // 2):
        if not batch:
            continue
        conditions = " OR ".join("(sync_meta_id = ? AND kind = ?)" for _ in batch)
        params = tuple(param for i, k in batch for param in (int(i), k.value))
        _DbState.connection().execute(
            f"DELETE FROM resource_file WHERE {conditions}", params
        )


def upsert_resource_files(params: Iterable[ResourceFileParams]) -> None:
    stmt = _SqlCache.get("upsert_resource_file.sql")
    _DbState.connection().executemany(stmt, (p.__dict__ for p in params))


# Discord webhook


def maybe_insert_discord_notification(song_id: SongId, resource: str) -> bool:
    """Does nothing if already present. Otherwise returns True."""
    stmt = (
        "INSERT OR IGNORE INTO discord_notification (song_id, resource) VALUES (?, ?)"
    )
    return _DbState.connection().execute(stmt, (song_id, resource)).rowcount > 0



"""Dataclasses and parse logic for meta tags from a txt's video tag."""

from __future__ import annotations

from typing import Literal

import attrs

from usdb_syncer.logger import Logger

# Characters that have special meaning for the meta tag syntax and therefore
# must be escaped. Escaping is done with percent encoding.
META_TAG_ESCAPES = ((",", "%2C"),)


def encode_meta_tag_value(meta_tag: str) -> str:
    """Escape special characters inside the value part of a meta tag."""
    for char, escape in META_TAG_ESCAPES:
        meta_tag = meta_tag.replace(char, escape)
    return meta_tag


def decode_meta_tag_value(meta_tag: str) -> str:
    """Unescape special characters inside the value part of a meta tag."""

    for char, escape in META_TAG_ESCAPES:
        meta_tag = meta_tag.replace(escape, char)
    return meta_tag


@attrs.define
class CropMetaTags:
    """Meta tags for cropping media."""

    left: int
    upper: int
    right: int
    lower: int

    @classmethod
    def try_parse(cls, value: str, logger: Logger) -> CropMetaTags | None:
        try:
            left, upper, width, height = map(int, value.split("-"))
        except ValueError:
            logger.warning(f"invalid value for crop meta tag: '{value}'")
            return None
        return cls(left, upper, left + width, upper + height)

    def to_str(self, prefix: str) -> str:
        width = self.right - self.left
        height = self.lower - self.upper
        value = f"{self.left}-{self.upper}-{width}-{height}"
        return _key_value_str(f"{prefix}-crop", value) or ""


@attrs.define
class ResizeMetaTags:
    """Meta tags for resizing media."""

    width: int
    height: int

    @classmethod
    def try_parse(cls, value: str, logger: Logger) -> ResizeMetaTags | None:
        try:
            if "-" in value:
                width, height = map(int, value.split("-"))
            else:
                width = height = int(value)
        except ValueError:
            logger.warning(f"invalid value for resize meta tag: '{value}'")
            return None
        return cls(width, height)

    def to_str(self, prefix: str) -> str:
        if self.width != self.height:
            value = f"{self.width}-{self.height}"
        else:
            value = str(self.width)
        return _key_value_str(f"{prefix}-resize", value) or ""


@attrs.define
class ImageMetaTags:
    """Meta tags relating to the cover or background image."""

    source: str
    rotate: float | None = None
    crop: CropMetaTags | None = None
    resize: ResizeMetaTags | None = None
    contrast: Literal["auto"] | float | None = None

    def source_url(self, logger: Logger) -> str:
        if "://" in self.source:
            if "fanart.tv" in self.source:
                logger.debug(
                    "Metatags contain a full fanart.tv URL instead of a fanart id only."
                )
                self.source = self.source.replace(
                    "images.fanart.tv", "assets.fanart.tv"
                )
            return self.source
        if "/" in self.source:
            logger.debug(f"{self.source} is missing the protocol.")
            return f"https://{self.source}"
        return f"https://assets.fanart.tv/fanart/{self.source}"

    def image_processing(self) -> bool:
        """True if there is data for image processing."""
        return any((self.rotate, self.crop, self.resize, self.contrast))

    def to_str(self, prefix: str) -> str:
        return _join_tags(
            _key_value_str(prefix, self.source),
            _key_value_str(f"{prefix}-rotate", self.rotate),
            self.crop.to_str(prefix) if self.crop else None,
            self.resize.to_str(prefix) if self.resize else None,
            _key_value_str(f"{prefix}-contrast", self.contrast),
        )


@attrs.define
class MedleyTag:
    """A tag with medley start and end."""

    start: int
    end: int

    @classmethod
    def try_parse(cls, value: str, logger: Logger) -> MedleyTag | None:
        try:
            start, end = map(int, value.split("-"))
        except ValueError:
            logger.warning(f"invalid value for medley meta tag: '{value}'")
            return None
        return cls(start, end)

    def __str__(self) -> str:
        return _key_value_str("medley", f"{self.start}-{self.end}") or ""


@attrs.define
class MetaTags:
    """Additional resource parameters from an overloaded video tag. Such an overloaded
    tag could look like:
    #VIDEO:a=example,co=foobar.jpg,bg=background.jpg
    """

    audio: str | None = None
    video: str | None = None
    cover: ImageMetaTags | None = None
    background: ImageMetaTags | None = None
    player1: str | None = None
    player2: str | None = None
    preview: float | None = None
    medley: MedleyTag | None = None
    tags: str | None = None

    @classmethod
    def parse(cls, video_tag: str, logger: Logger) -> MetaTags:
        meta_tags = cls()
        if "=" not in video_tag:
            # probably a regular video file name and not a meta tag
            return meta_tags
        for pair in video_tag.split(","):
            if "=" not in pair:
                logger.warning(f"missing key or value for meta tag: '{pair}'")
                continue
            key, value = pair.split("=", maxsplit=1)
            meta_tags._parse_key_value_pair(key, value, logger)
        return meta_tags

    def _parse_key_value_pair(  # noqa: C901
        self, key: str, value: str, logger: Logger
    ) -> None:
        value = decode_meta_tag_value(value)
        match key:
            case "v":
                self.video = value
            case "v-trim" | "v-crop" | "co-protocol" | "bg-protocol":
                logger.debug(f"Unused meta tag found: {key}={value}")
            case "a":
                self.audio = value
            case "co":
                self.cover = ImageMetaTags(source=value)
            case "co-rotate" if self.cover:
                self.cover.rotate = _try_parse_float(value, logger)
            case "co-crop" if self.cover:
                self.cover.crop = CropMetaTags.try_parse(value, logger)
            case "co-resize" if self.cover:
                self.cover.resize = ResizeMetaTags.try_parse(value, logger)
            case "co-contrast" if self.cover:
                self.cover.contrast = _try_parse_contrast(value, logger)
            case "bg":
                self.background = ImageMetaTags(source=value)
            case "bg-crop" if self.background:
                self.background.crop = CropMetaTags.try_parse(value, logger)
            case "bg-resize" if self.background:
                self.background.resize = ResizeMetaTags.try_parse(value, logger)
            case "p1":
                self.player1 = value
            case "p2":
                self.player2 = value
            case "preview":
                self.preview = _try_parse_float(value, logger)
            case "medley":
                self.medley = MedleyTag.try_parse(value, logger)
            case "tags":
                self.tags = value
            case _:
                logger.warning(f"unknown key for meta tag: '{key}={value}'")

    def is_audio_only(self) -> bool:
        """True if a resource is explicitly set for audio only."""
        return bool(self.audio and not self.video)

    def __str__(self) -> str:
        return _join_tags(
            _key_value_str("a", self.audio),
            _key_value_str("v", self.video),
            self.cover.to_str("co") if self.cover else None,
            self.background.to_str("bg") if self.background else None,
            _key_value_str("p1", self.player1),
            _key_value_str("p2", self.player2),
            _key_value_str("preview", self.preview),
            str(self.medley) if self.medley else None,
            _key_value_str("tags", self.tags),
        )


def _key_value_str(key: str, value: str | float | None) -> str | None:
    return None if value is None else f"{key}={encode_meta_tag_value(str(value))}"


def _join_tags(*meta_tags: str | None) -> str:
    return ",".join(filter(None, meta_tags))


def _try_parse_float(value: str, logger: Logger) -> float | None:
    try:
        return float(value)
    except ValueError:
        logger.warning(f"invalid number for meta tag: '{value}'")
        return None


def _try_parse_contrast(value: str, logger: Logger) -> Literal["auto"] | float | None:
    if value == "auto":
        return "auto"
    try:
        return float(value)
    except ValueError:
        logger.warning(f"invalid value for contrast meta tag: '{value}'")
    return None



"""Debug Console with a REPL."""

import io
import traceback
from contextlib import redirect_stdout

from PySide6.QtGui import QFontDatabase, QTextCursor
from PySide6.QtWidgets import QDialog, QWidget

from usdb_syncer.gui import gui_utils
from usdb_syncer.gui.forms.DebugConsole import Ui_Dialog


class DebugConsole(Ui_Dialog, QDialog):
    """Debug Console with a REPL."""

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self._set_font()
        self._set_shortcuts()

    def _set_font(self) -> None:
        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        self.input.setFont(font)
        self.output.setFont(font)

    def _set_shortcuts(self) -> None:
        gui_utils.set_shortcut("Ctrl+Return", self, self._execute)
        gui_utils.set_shortcut("Ctrl+Shift+Return", self, self._execute_and_print)
        gui_utils.set_shortcut("Ctrl+L", self, self.output.clear)
        gui_utils.set_shortcut("Ctrl+Shift+L", self, self.input.clear)

    def _execute_and_print(self) -> None:
        cursor = self.input.textCursor()
        position = cursor.position()
        cursor.select(QTextCursor.SelectionType.LineUnderCursor)
        line = cursor.selectedText()
        whitespace, stripped = split_off_leading_whitespace(line)
        opening, closing = "print(", ")"
        if not stripped.startswith(opening):
            line = f"{whitespace}{opening}{stripped}{closing}"
            cursor.insertText(line)
            cursor.setPosition(position + len(opening))
            self.input.setTextCursor(cursor)
        self._execute()

    def _execute(self) -> None:
        code = self.input.toPlainText()
        with redirect_stdout(io.StringIO()) as captured:
            try:
                exec(code, {"mw": self.parent()})  # noqa: S102
            except Exception:  # noqa: BLE001
                print(traceback.format_exc())
        self._log_output(code, captured.getvalue())

    def _log_output(self, code: str, output: str) -> None:
        indented_code = ">>> " + code.strip().replace("\n", "\n... ") + "\n"
        output = output or "<no output>"
        self.output.appendPlainText(indented_code + (output or "<no output>"))
        gui_utils.scroll_to_bottom(self.output)


def split_off_leading_whitespace(text: str) -> tuple[str, str]:
    stripped = text.lstrip()
    whitespace = text[: len(text) - len(stripped)]
    return whitespace, stripped



"""A download manager for USDB songs."""

from __future__ import annotations

import base64
import binascii
import random
from pathlib import Path

from usdb_syncer import errors
from usdb_syncer._version import __version__ as __version__
from usdb_syncer.constants import Usdb


class SongId(int):
    """Bounded int representing an id on USDB.

    Use builtin str to get the canonical zero-padded five-digit str.
    """

    def __init__(self, value: int) -> None:
        if value not in range(100_000):
            raise errors.SongIdError(value)

    def __str__(self) -> str:
        return f"{self:05}"

    @classmethod
    def parse(cls, value: str) -> SongId:
        return cls(int(value))

    @classmethod
    def try_parse(cls, value: str) -> SongId | None:
        try:
            return cls.parse(value)
        except ValueError:
            return None

    def usdb_gettxt_url(self) -> str:
        return f"{Usdb.GETTXT_URL}{self:d}"

    def usdb_detail_url(self) -> str:
        return f"{Usdb.DETAIL_URL}{self:d}"


class SyncMetaId(int):
    """8-byte signed integer with str encoding."""

    _len_bytes = 8
    # chars needed to base64-encode _len_bytes bytes without padding
    _len_chars = 11

    @classmethod
    def new(cls) -> SyncMetaId:
        return cls(random.randint(-(2**63), 2**63 - 1))  # noqa: S311

    def encode(self) -> str:
        value = base64.urlsafe_b64encode(
            self.to_bytes(self._len_bytes, "big", signed=True)
        ).decode()
        # strip padding
        return value[:-1]

    @classmethod
    def decode(cls, value: str) -> SyncMetaId | None:
        if len(value) != cls._len_chars:
            return None
        try:
            number = base64.urlsafe_b64decode(f"{value}=")
        except (binascii.Error, ValueError):
            return None
        if len(number) != cls._len_bytes:
            return None
        return cls.from_bytes(number, "big", signed=True)

    @classmethod
    def from_path(cls, path: Path) -> SyncMetaId | None:
        return cls.decode(path.stem)

    def to_filename(self) -> str:
        return f"{self.encode()}.usdb"



"""Dialog with app settings."""

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QDialog, QWidget

import usdb_syncer
from usdb_syncer.gui.forms.AboutDialog import Ui_Dialog

RESET_GAP = 4000
SCROLL_STEP = 1
SCROLL_GAP = 100


class AboutDialog(Ui_Dialog, QDialog):
    """Dialog with about info and credits."""

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self.label_version.setText(usdb_syncer.__version__)
        self._reset_text()

    def _scroll_down(self) -> None:
        scroll_bar = self.credits.verticalScrollBar()
        if scroll_bar.value() >= scroll_bar.maximum():
            QTimer.singleShot(RESET_GAP, self._reset_text)
        else:
            scroll_bar.setValue(scroll_bar.value() + SCROLL_STEP)
            QTimer.singleShot(SCROLL_GAP, self._scroll_down)

    def _reset_text(self) -> None:
        scroll_bar = self.credits.verticalScrollBar()
        scroll_bar.setValue(scroll_bar.minimum())
        QTimer.singleShot(SCROLL_GAP, self._scroll_down)



"""Table model for song data."""

from collections.abc import Iterable
from functools import cache
from typing import Any, assert_never

from PySide6.QtCore import (
    QAbstractTableModel,
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    Qt,
)
from PySide6.QtGui import QIcon

from usdb_syncer import SongId, events, utils
from usdb_syncer.gui import icons
from usdb_syncer.gui.song_table.column import Column
from usdb_syncer.usdb_song import DownloadStatus, UsdbSong

QIndex = QModelIndex | QPersistentModelIndex


class TableModel(QAbstractTableModel):
    """Table model for song data."""

    _ids: tuple[SongId, ...] = ()
    _rows: dict[SongId, int]

    def __init__(self, parent: QObject) -> None:
        super().__init__(parent)
        self._rows = {}
        events.SongChanged.subscribe(self._on_song_changed)
        events.SongDeleted.subscribe(self._on_song_deleted)
        events.SongDirChanged.subscribe(lambda _: self.reset)

    def reset(self) -> None:
        self.beginResetModel()
        self.endResetModel()

    def set_songs(self, songs: Iterable[SongId]) -> None:
        self.beginResetModel()
        self._ids = tuple(songs)
        self._rows = {song: row for row, song in enumerate(self._ids)}
        self.endResetModel()

    def ids_for_indices(self, indices: Iterable[QModelIndex]) -> list[SongId]:
        return [self._ids[idx.row()] for idx in indices]

    def ids_for_rows(self, rows: Iterable[int]) -> list[SongId]:
        return [self._ids[row] for row in rows]

    def indices_for_ids(self, ids: Iterable[SongId]) -> list[QModelIndex]:
        return [
            self.index(row, 0)
            for song_id in ids
            if (row := self._rows.get(song_id)) is not None
        ]

    def row_for_id(self, song_id: SongId) -> int | None:
        return self._rows.get(song_id)

    def _row_changed(self, row: int) -> None:
        start_idx = self.index(row, 0)
        end_idx = self.index(row, self.columnCount() - 1)
        self.dataChanged.emit(start_idx, end_idx)

    def _on_song_changed(self, event: events.SongChanged) -> None:
        if (row := self._rows.get(event.song_id)) is not None:
            self._row_changed(row)

    def _on_song_deleted(self, event: events.SongDeleted) -> None:
        if (row := self._rows.get(event.song_id)) is None:
            return
        self.beginRemoveRows(QModelIndex(), row, row)
        del self._rows[event.song_id]
        self._ids = self._ids[:row] + self._ids[row + 1 :]
        self.endRemoveRows()

    # QAbstractTableModel implementation

    def columnCount(self, parent: QIndex | None = None) -> int:  # noqa: N802
        if parent is None:
            parent = QModelIndex()
        return 0 if parent.isValid() else len(Column)

    def rowCount(self, parent: QIndex | None = None) -> int:  # noqa: N802
        if parent is None:
            parent = QModelIndex()
        return 0 if parent.isValid() else len(self._ids)

    def data(self, index: QIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            if song := self._get_song(index):
                return _display_data(song, index.column())
        if role == Qt.ItemDataRole.DecorationRole:
            if song := self._get_song(index):
                return _decoration_data(song, index.column())
        return None

    def _get_song(self, index: QIndex) -> UsdbSong | None:
        if not index.isValid():
            return None
        song_id = self._ids[index.row()]
        if song := UsdbSong.get(song_id):
            return song
        return None

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation != Qt.Orientation.Horizontal:
            return None
        if role == Qt.ItemDataRole.DecorationRole:
            return Column(section).decoration_data()
        if role == Qt.ItemDataRole.DisplayRole:
            return Column(section).display_data()
        return None


def _display_data(song: UsdbSong, column: int) -> str | None:  # noqa: C901
    col = Column(column)
    match col:
        case Column.SONG_ID:
            return str(song.song_id)
        case Column.ARTIST:
            return song.artist
        case Column.TITLE:
            return song.title
        case Column.LANGUAGE:
            return song.language
        case Column.EDITION:
            return song.edition
        case Column.GOLDEN_NOTES:
            return yes_no_str(song.golden_notes)
        case Column.RATING:
            return rating_str(song.rating)
        case Column.VIEWS:
            return str(song.views)
        case Column.YEAR:
            return str(song.year) if song.year else ""
        case Column.GENRE:
            return song.genre
        case Column.CREATOR:
            return song.creator
        case Column.TAGS:
            return song.tags
        case Column.DOWNLOAD_STATUS:
            return (
                utils.format_timestamp(song.sync_meta.mtime)
                if song.sync_meta and song.status is DownloadStatus.NONE
                else str(song.status)
            )
        case (
            Column.SAMPLE_URL
            | Column.TXT
            | Column.AUDIO
            | Column.VIDEO
            | Column.COVER
            | Column.BACKGROUND
            | Column.PINNED
        ):
            return None
        case _ as unreachable:
            assert_never(unreachable)


def _decoration_data(song: UsdbSong, column: int) -> QIcon | None:  # noqa: C901
    col = Column(column)
    match col:
        case (
            Column.SONG_ID
            | Column.ARTIST
            | Column.TITLE
            | Column.LANGUAGE
            | Column.EDITION
            | Column.GOLDEN_NOTES
            | Column.RATING
            | Column.VIEWS
            | Column.DOWNLOAD_STATUS
            | Column.YEAR
            | Column.GENRE
            | Column.CREATOR
            | Column.TAGS
        ):
            return None
        case Column.SAMPLE_URL:
            local = bool(song.sync_meta and song.sync_meta.audio)
            if song.is_playing and local:
                icon = icons.Icon.PAUSE_LOCAL
            elif song.is_playing:
                icon = icons.Icon.PAUSE_REMOTE
            elif local:
                icon = icons.Icon.PLAY_LOCAL
            else:
                icon = icons.Icon.PLAY_REMOTE
        case Column.TXT:
            if not (song.sync_meta and song.sync_meta.txt):
                return None
            icon = icons.Icon.CHECK
        case Column.AUDIO:
            if not (song.sync_meta and song.sync_meta.audio):
                return None
            icon = icons.Icon.CHECK
        case Column.VIDEO:
            if not (song.sync_meta and song.sync_meta.video):
                return None
            icon = icons.Icon.CHECK
        case Column.COVER:
            if not (song.sync_meta and song.sync_meta.cover):
                return None
            icon = icons.Icon.CHECK
        case Column.BACKGROUND:
            if not (song.sync_meta and song.sync_meta.background):
                return None
            icon = icons.Icon.CHECK
        case Column.PINNED:
            if not (song.sync_meta and song.sync_meta.pinned):
                return None
            icon = icons.Icon.PIN
        case _ as unreachable:
            assert_never(unreachable)
    return icon.icon()


@cache
def rating_str(rating: int) -> str:
    return rating * "★"


def yes_no_str(yes: bool) -> str:
    return "Yes" if yes else "No"



"""Functions for postprocessing audio/video files."""

from base64 import b64encode
from os import devnull
from pathlib import Path

from ffmpeg_normalize import FFmpegNormalize
from mutagen import id3
from mutagen.flac import Picture
from mutagen.id3 import ID3, TXXX
from mutagen.mp3 import MP3
from mutagen.mp4 import MP4, MP4Cover
from mutagen.ogg import OggFileType
from mutagen.oggopus import OggOpus
from mutagen.oggvorbis import OggVorbis
from PIL import Image

from usdb_syncer import settings
from usdb_syncer.constants import ISO_639_2B_LANGUAGE_CODES
from usdb_syncer.download_options import AudioOptions, VideoOptions
from usdb_syncer.logger import Logger
from usdb_syncer.settings import AudioFormat, AudioNormalization, VideoContainer
from usdb_syncer.song_txt import SongTxt

# Normalization parameters
DEFAULT_TARGET_LEVEL_RG_DB = -18.0
DEFAULT_TARGET_LEVEL_R128_DB = -23.0

# Explicitly initialize the Python Image Library
Image.init()


def normalize_audio(
    options: AudioOptions, path_stem: Path, input_ext: str, logger: Logger
) -> None:
    normalizer = _create_normalizer(options)
    input_file = f"{path_stem}.{input_ext}"
    output_file = f"{path_stem}.{options.format.value}"
    if options.normalization == AudioNormalization.REPLAYGAIN:
        # audio file is already in target format from postprocessor
        input_file = f"{path_stem}.{options.format.value}"
        # we do not want to actually rewrite the file, so we use 'null'
        output_file = devnull

    normalizer.add_media_file(input_file, output_file)
    normalizer.run_normalization()

    if options.normalization == AudioNormalization.REPLAYGAIN:
        _write_replaygain_tags(normalizer, input_file, logger)


def _create_normalizer(options: AudioOptions) -> FFmpegNormalize:
    target_level = (
        DEFAULT_TARGET_LEVEL_R128_DB
        if options.format == settings.AudioFormat.OPUS
        else DEFAULT_TARGET_LEVEL_RG_DB
    )
    normalizer = FFmpegNormalize(
        normalization_type="ebu",
        target_level=target_level,
        keep_loudness_range_target=True,
        dynamic=False,
        audio_codec=options.format.ffmpeg_encoder(),
        audio_bitrate=options.bitrate.ffmpeg_format(),
        progress=True,
        extra_output_options=(
            ["-f", "null"]
            if options.normalization == AudioNormalization.REPLAYGAIN
            else []
        ),
    )

    return normalizer


def _write_replaygain_tags(
    normalizer: FFmpegNormalize, audio_file: str, logger: Logger
) -> None:
    """Writes ReplayGain values to audio file metadata."""

    # extract stats from 2-pass normalization, then set replay gain values to audio file
    stats = next(iter(normalizer.media_files[0].get_stats()), None)
    if not stats:
        logger.warning("Normalization failed: no stats")
        return
    ebu_pass2 = stats.get("ebu_pass2")
    if not ebu_pass2:
        logger.warning("Normalization failed: no EBU pass 2 normalization stats")
        return

    if (input_i := ebu_pass2.get("input_i")) is None or (
        input_tp := ebu_pass2.get("input_tp")
    ) is None:
        logger.warning("Normalization failed: no loudness or true peak stats")
        return
    track_gain = normalizer.target_level - input_i  # dB
    track_peak = 10 ** (input_tp / 20)  # Linear scale

    try:
        match Path(audio_file).suffix[1:]:
            case AudioFormat.M4A.value:
                _write_replaygain_tags_m4a(audio_file, track_gain, track_peak)
            case AudioFormat.MP3.value:
                _write_replaygain_tags_mp3(audio_file, track_gain, track_peak)
            case AudioFormat.OGG.value:
                _write_replaygain_tags_ogg(audio_file, track_gain, track_peak)
            case AudioFormat.OPUS.value:
                _write_replaygain_tags_opus(audio_file, track_gain)
            case other:
                logger.debug(f"ReplayGain tags not supported for suffix '{other}'.")

        logger.info(f"ReplayGain tags written to {audio_file}")

    except Exception:
        logger.exception(f"Failed to write audio tags to file '{audio_file}'!")
    else:
        logger.debug(f"Audio tags written to file '{audio_file}'.")


def _write_replaygain_tags_m4a(
    audio_file: str, track_gain: float, track_peak: float
) -> None:
    mp4 = MP4(audio_file)
    if not mp4.tags:
        mp4.add_tags()
    if not mp4.tags:
        return
    mp4.tags["----:com.apple.iTunes:REPLAYGAIN_TRACK_GAIN"] = [
        f"{track_gain:.2f} dB".encode()
    ]
    mp4.tags["----:com.apple.iTunes:REPLAYGAIN_TRACK_PEAK"] = [
        f"{track_peak:.6f}".encode()
    ]
    mp4.save()


def _write_replaygain_tags_mp3(
    audio_file: str, track_gain: float, track_peak: float
) -> None:
    mp3 = MP3(audio_file, ID3=ID3)
    if not mp3.tags:
        return
    mp3.tags.add(TXXX(desc="REPLAYGAIN_TRACK_GAIN", text=[f"{track_gain:.2f} dB"]))
    mp3.tags.add(TXXX(desc="REPLAYGAIN_TRACK_PEAK", text=[f"{track_peak:.6f}"]))
    mp3.save()


def _write_replaygain_tags_ogg(
    audio_file: str, track_gain: float, track_peak: float
) -> None:
    ogg = OggVorbis(audio_file)
    ogg["REPLAYGAIN_TRACK_GAIN"] = [f"{track_gain:.2f} dB"]
    ogg["REPLAYGAIN_TRACK_PEAK"] = [f"{track_peak:.6f}"]
    ogg.save()


def _write_replaygain_tags_opus(audio_file: str, track_gain: float) -> None:
    opus = OggOpus(audio_file)
    # See https://datatracker.ietf.org/doc/html/rfc7845#section-5.2.1
    opus["R128_TRACK_GAIN"] = [str(round(256 * track_gain))]
    opus.save()


def write_audio_tags(
    *,
    txt: SongTxt,
    options: AudioOptions,
    audio: tuple[Path, str],
    cover: tuple[Path, str] | None,
    background: tuple[Path, str] | None,
    logger: Logger,
) -> None:
    audio_path = audio[0]
    try:
        match audio_path.suffix[1:]:
            case AudioFormat.M4A.value:
                _write_tags_m4a_mp4(
                    audio, cover, background, txt, options.embed_artwork
                )
            case AudioFormat.MP3.value:
                _write_tags_mp3(audio, cover, txt, options.embed_artwork)
            case AudioFormat.OGG.value:
                _write_tags_ogg(audio, cover, txt, options.embed_artwork)
            case AudioFormat.OPUS.value:
                _write_tags_ogg(audio, cover, txt, options.embed_artwork)
            case other:
                logger.debug(f"Audio tags not supported for suffix '{other}'.")
                return
    except Exception:
        logger.exception(f"Failed to write audio tags to file '{audio_path}'!")
    else:
        logger.debug(f"Audio tags written to file '{audio_path}'.")


def write_video_tags(
    *,
    txt: SongTxt,
    options: VideoOptions,
    video: tuple[Path, str],
    cover: tuple[Path, str] | None,
    background: tuple[Path, str] | None,
    logger: Logger,
) -> None:
    video_path = video[0]
    try:
        match video_path.suffix[1:]:
            case VideoContainer.MP4.value:
                _write_tags_m4a_mp4(
                    video, cover, background, txt, options.embed_artwork
                )
            case other:
                logger.debug(f"Video tags not supported for suffix '{other}'.")
                return
    except Exception:
        logger.exception(f"Failed to write video tags to file '{video_path}'!")
    else:
        logger.debug(f"Video tags written to file '{video_path}'.")


def _write_tags_m4a_mp4(
    audio_video: tuple[Path, str],
    cover: tuple[Path, str] | None,
    background: tuple[Path, str] | None,
    txt: SongTxt,
    embed_artwork: bool,
) -> None:
    m4a_mp4 = MP4(audio_video[0])
    if not m4a_mp4.tags:
        m4a_mp4.add_tags()
    if not m4a_mp4.tags:
        return
    m4a_mp4.tags["\xa9ART"] = txt.headers.artist
    m4a_mp4.tags["\xa9nam"] = txt.headers.title
    if txt.headers.genre:
        m4a_mp4.tags["\xa9gen"] = txt.headers.genre
    if txt.headers.year:
        m4a_mp4.tags["\xa9day"] = txt.headers.year
    m4a_mp4.tags["\xa9lyr"] = txt.unsynchronized_lyrics()
    m4a_mp4.tags["\xa9cmt"] = audio_video[1]

    if embed_artwork:
        cover_path = cover[0] if cover else None
        background_path = background[0] if background else None
        m4a_mp4.tags["covr"] = [
            MP4Cover(image.read_bytes(), imageformat=MP4Cover.FORMAT_JPEG)
            for image in (cover_path, background_path)
            if image
        ]

    m4a_mp4.save()


def _write_tags_mp3(
    audio: tuple[Path, str],
    cover: tuple[Path, str] | None,
    txt: SongTxt,
    embed_artwork: bool,
) -> None:
    audio_path, audio_resource = audio
    mp3 = MP3(audio_path, ID3=id3.ID3)
    if not mp3.tags:
        mp3.add_tags()
    if not mp3.tags:
        return
    lang = ISO_639_2B_LANGUAGE_CODES.get(txt.headers.main_language(), "und")
    mp3.tags["TPE1"] = id3.TPE1(encoding=id3.Encoding.UTF8, text=txt.headers.artist)
    mp3.tags["TIT2"] = id3.TIT2(encoding=id3.Encoding.UTF8, text=txt.headers.title)
    mp3.tags["TLAN"] = id3.TLAN(encoding=id3.Encoding.UTF8, text=lang)
    if genre := txt.headers.genre:
        mp3.tags["TCON"] = id3.TCON(encoding=id3.Encoding.UTF8, text=genre)
    if year := txt.headers.year:
        mp3.tags["TDRC"] = id3.TDRC(encoding=id3.Encoding.UTF8, text=year)
    mp3.tags[f"USLT::'{lang}'"] = id3.USLT(
        encoding=id3.Encoding.UTF8,
        lang=lang,
        desc="Lyrics",
        text=txt.unsynchronized_lyrics(),
    )
    mp3.tags["SYLT"] = id3.SYLT(
        encoding=id3.Encoding.UTF8,
        lang=lang,
        format=2,  # milliseconds as units
        type=1,  # lyrics
        text=txt.synchronized_lyrics(),
    )
    mp3.tags["COMM"] = id3.COMM(
        encoding=id3.Encoding.UTF8, lang="eng", desc="Audio Source", text=audio_resource
    )

    if embed_artwork and cover:
        cover_path, cover_resource = cover
        mp3.tags.add(
            id3.APIC(
                encoding=id3.Encoding.UTF8,
                mime=Image.MIME["JPEG"],
                type=id3.PictureType.COVER_FRONT,
                desc=f"Source: {cover_resource}",
                data=cover_path.read_bytes(),
            )
        )

    mp3.save()


def _write_tags_ogg(
    audio: tuple[Path, str],
    cover: tuple[Path, str] | None,
    txt: SongTxt,
    embed_artwork: bool,
) -> None:
    audio_path, audio_resource = audio
    ogg = OggFileType()
    match audio_path.suffix[1:]:
        case AudioFormat.OGG.value:
            ogg = OggVorbis(audio_path)
        case AudioFormat.OPUS.value:
            ogg = OggOpus(audio_path)
        case _:
            return
    # Set basic tags
    ogg["artist"] = txt.headers.artist
    ogg["title"] = txt.headers.title
    ogg["language"] = ISO_639_2B_LANGUAGE_CODES.get(txt.headers.main_language(), "und")
    if genre := txt.headers.genre:
        ogg["genre"] = genre
    if year := txt.headers.year:
        ogg["date"] = year
    ogg["lyrics"] = txt.unsynchronized_lyrics()
    ogg["comment"] = audio_resource

    if embed_artwork and cover:
        cover_path = cover[0]
        picture = Picture()
        with cover_path.open("rb") as file:
            picture.data = file.read()
        with Image.open(cover_path) as image:
            picture.width, picture.height = image.size
        picture.type = 3  # "Cover (front)"
        picture.desc = "Cover art"
        picture.mime = Image.MIME["JPEG"]
        picture.depth = 24

        encoded_data = b64encode(picture.write())
        vcomment_value = encoded_data.decode("ascii")

        ogg["metadata_block_picture"] = [vcomment_value]

    ogg.save()



"""Generates a JSON from the passed song list."""

from __future__ import annotations

import datetime
import json
from collections.abc import Iterable
from json import JSONEncoder
from pathlib import Path
from typing import Any

import attrs

from usdb_syncer import SongId
from usdb_syncer.logger import logger
from usdb_syncer.usdb_song import UsdbSong
from usdb_syncer.utils import video_url_from_resource

JSON_EXPORT_VERSION = 1


@attrs.define(kw_only=True)
class SongExportData:
    """Meta Data describing a song from USDB including specific meta tag data"""

    id: SongId
    artist: str
    title: str
    year: int | None = None
    edition: str | None = None
    genre: str | None = None
    tags: str | None = None
    language: str | None = None
    golden_notes: bool
    cover_url: str | None = None
    cover_meta: str | None = None
    audio_url: str | None = None
    video_url: str | None = None
    duet: bool

    @classmethod
    def from_usdb_song(cls, song: UsdbSong) -> SongExportData | None:
        if not (meta := song.sync_meta):
            return None
        return cls(
            id=song.song_id,
            artist=song.artist,
            title=song.title,
            year=song.year,
            edition=(
                None if (not song.edition or song.edition == "None") else song.edition
            ),
            genre=song.genre,
            tags=song.tags,
            language=song.language,
            golden_notes=song.golden_notes,
            cover_url=(
                meta.meta_tags.cover.source_url(logger)
                if meta.meta_tags.cover
                else None
            ),
            cover_meta=(
                meta.meta_tags.cover.to_str("co") if meta.meta_tags.cover else None
            ),
            audio_url=(
                video_url_from_resource(meta.meta_tags.audio)
                if meta.meta_tags.audio
                else None
            ),
            video_url=(
                video_url_from_resource(meta.meta_tags.video)
                if meta.meta_tags.video
                else None
            ),
            duet=(
                meta.meta_tags.player1 is not None
                and meta.meta_tags.player2 is not None
            ),
        )


@attrs.define(kw_only=True)
class JsonSongList:
    """defines fields in JSON songlist export"""

    songs: list[SongExportData]
    date: str
    version: int = attrs.field(default=JSON_EXPORT_VERSION, init=False)

    @classmethod
    def from_songs(
        cls, songs: Iterable[SongId], date: datetime.datetime
    ) -> JsonSongList:
        song_list = [
            song_data
            for song_id in songs
            if (song := UsdbSong.get(song_id))
            and (song_data := SongExportData.from_usdb_song(song))
        ]
        return cls(songs=song_list, date=str(date))


class JsonSongListEncoder(JSONEncoder):
    """Custom JSON encoder for SongExportData"""

    def default(self, o: Any) -> Any:
        if isinstance(o, JsonSongList):
            dct = attrs.asdict(o, recurse=True)
            return dct
        return super().default(o)


def generate_report_json(
    songs: Iterable[SongId], path: Path, indent: int = 4
) -> tuple[Path, int]:
    content = JsonSongList.from_songs(songs=songs, date=datetime.datetime.now())
    with path.open("w", encoding="utf8") as file:
        json.dump(
            content, file, cls=JsonSongListEncoder, indent=indent, ensure_ascii=False
        )
    return path, len(content.songs)



"""Dialog with app settings."""

import sys
from pathlib import Path
from typing import assert_never

from PySide6 import QtWidgets
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QFileDialog, QWidget

from usdb_syncer import SongId, path_template, settings
from usdb_syncer.gui import icons, theme
from usdb_syncer.gui.forms.SettingsDialog import Ui_Dialog
from usdb_syncer.path_template import PathTemplate
from usdb_syncer.usdb_scraper import SessionManager
from usdb_syncer.usdb_song import UsdbSong

_FALLBACK_SONG = UsdbSong(
    song_id=SongId(3715),
    artist="Queen",
    title="Bohemian Rhapsody",
    genre="Genre",
    year=1975,
    creator="IC",
    edition="[SC]-Songs",
    language="English",
    golden_notes=True,
    rating=5,
    views=4050,
    sample_url="",
)


class SettingsDialog(Ui_Dialog, QDialog):
    """Dialog with app settings."""

    _path_template: PathTemplate | None = None

    def __init__(self, parent: QWidget, song: UsdbSong | None) -> None:
        super().__init__(parent=parent)
        self._song = song or _FALLBACK_SONG
        self.setupUi(self)
        self._populate_comboboxes()
        self._load_settings()
        self._setup_path_template()
        apply_btn = self.buttonBox.button(QDialogButtonBox.StandardButton.Apply)
        apply_btn.clicked.connect(self.apply)
        self._browser = self.comboBox_browser.currentData()
        self.groupBox_reencode_video.setVisible(False)
        if sys.platform != "win32":
            self.groupBox_vocaluxe.setVisible(False)
        self.pushButton_browse_karedi.clicked.connect(
            lambda: self._set_location(settings.SupportedApps.KAREDI)
        )
        self.pushButton_browse_performous.clicked.connect(
            lambda: self._set_location(settings.SupportedApps.PERFORMOUS)
        )
        self.pushButton_browse_ultrastar_manager.clicked.connect(
            lambda: self._set_location(settings.SupportedApps.ULTRASTAR_MANAGER)
        )
        self.pushButton_browse_usdx.clicked.connect(
            lambda: self._set_location(settings.SupportedApps.USDX)
        )
        self.pushButton_browse_vocaluxe.clicked.connect(
            lambda: self._set_location(settings.SupportedApps.VOCALUXE)
        )
        self.pushButton_browse_yass_reloaded.clicked.connect(
            lambda: self._set_location(settings.SupportedApps.YASS_RELOADED)
        )
        self.comboBox_theme.currentIndexChanged.connect(
            self._set_theme_settings_enabled
        )
        self._set_theme_settings_enabled()

    def _set_theme_settings_enabled(self) -> None:
        hidden = self.comboBox_theme.currentData() == settings.Theme.SYSTEM
        self.label_primary_color.setHidden(hidden)
        self.comboBox_primary_color.setHidden(hidden)
        self.checkBox_colored_background.setHidden(hidden)

    def _set_location(self, app: settings.SupportedApps) -> None:
        path = self._get_executable(app)
        text = str(path) if path is not None else ""
        match app:
            case settings.SupportedApps.KAREDI:
                self.lineEdit_path_karedi.setText(text)
            case settings.SupportedApps.PERFORMOUS:
                self.lineEdit_path_performous.setText(text)
            case settings.SupportedApps.ULTRASTAR_MANAGER:
                self.lineEdit_path_ultrastar_manager.setText(text)
            case settings.SupportedApps.USDX:
                self.lineEdit_path_usdx.setText(text)
            case settings.SupportedApps.VOCALUXE:
                self.lineEdit_path_vocaluxe.setText(text)
            case settings.SupportedApps.YASS_RELOADED:
                self.lineEdit_path_yass_reloaded.setText(text)
            case _ as unreachable:
                assert_never(unreachable)

    def _get_executable(self, app: settings.SupportedApps) -> Path | None:
        filt = "*"
        directory = ""
        match sys.platform:
            case "win32":
                filt = "*.exe"
            case "darwin":
                directory = "/Applications"
                filt = "*.app"
        if app in (settings.SupportedApps.KAREDI, settings.SupportedApps.YASS_RELOADED):
            filt += " *.jar"
        filename = QFileDialog.getOpenFileName(
            self, f"Select {app} executable", directory, filt
        )[0]
        # dialog is hidden by main window on macOS if file picker was cancelled
        self.raise_()
        if filename == "":
            return None
        path = Path(filename)
        if path.suffix != ".app":
            return path
        if (
            full_path := path.joinpath("Contents", "MacOS", app.executable_name())
        ).exists():
            return full_path
        return None

    def _populate_comboboxes(self) -> None:
        combobox_settings = (
            (self.comboBox_theme, settings.Theme),
            (self.comboBox_primary_color, settings.Color),
            (self.comboBox_encoding, settings.Encoding),
            (self.comboBox_line_endings, settings.Newline),
            (self.comboBox_format_version, settings.FormatVersion),
            (self.comboBox_fix_linebreaks, settings.FixLinebreaks),
            (self.comboBox_fix_spaces, settings.FixSpaces),
            (self.comboBox_cover_max_size, settings.CoverMaxSize),
            (self.comboBox_ytdlp_rate_limit, settings.YtdlpRateLimit),
            (self.comboBox_audio_format, settings.AudioFormat),
            (self.comboBox_audio_bitrate, settings.AudioBitrate),
            (self.comboBox_audio_normalization, settings.AudioNormalization),
            (self.comboBox_videocontainer, settings.VideoContainer),
            (self.comboBox_videoencoder, settings.VideoCodec),
            (self.comboBox_videoresolution, settings.VideoResolution),
            (self.comboBox_fps, settings.VideoFps),
        )
        for combobox, setting in combobox_settings:
            for item in setting:
                combobox.addItem(str(item), item)
        for browser in settings.Browser:
            if icon := icons.browser_icon(browser):
                self.comboBox_browser.addItem(icon, str(browser), browser)
            else:
                self.comboBox_browser.addItem(str(browser), browser)

    def _load_settings(self) -> None:
        self.comboBox_theme.setCurrentIndex(
            self.comboBox_theme.findData(settings.get_theme())
        )
        self.comboBox_primary_color.setCurrentIndex(
            self.comboBox_primary_color.findData(settings.get_primary_color())
        )
        self.checkBox_colored_background.setChecked(settings.get_colored_background())
        self.comboBox_browser.setCurrentIndex(
            self.comboBox_browser.findData(settings.get_browser())
        )
        self.groupBox_cover.setChecked(settings.get_cover())
        self.comboBox_cover_max_size.setCurrentIndex(
            self.comboBox_cover_max_size.findData(settings.get_cover_max_size())
        )
        self.groupBox_songfile.setChecked(settings.get_txt())
        self.comboBox_encoding.setCurrentIndex(
            self.comboBox_encoding.findData(settings.get_encoding())
        )
        self.comboBox_line_endings.setCurrentIndex(
            self.comboBox_line_endings.findData(settings.get_newline())
        )
        self.comboBox_format_version.setCurrentIndex(
            self.comboBox_format_version.findData(settings.get_version())
        )
        self.comboBox_fix_linebreaks.setCurrentIndex(
            self.comboBox_fix_linebreaks.findData(settings.get_fix_linebreaks())
        )
        self.checkBox_fix_first_words_capitalization.setChecked(
            settings.get_fix_first_words_capitalization()
        )
        self.comboBox_fix_spaces.setCurrentIndex(
            self.comboBox_fix_spaces.findData(settings.get_fix_spaces())
        )
        self.checkBox_fix_quotation_marks.setChecked(settings.get_fix_quotation_marks())
        self.spinBox_throttling_threads.setValue(settings.get_throttling_threads())
        self.comboBox_ytdlp_rate_limit.setCurrentIndex(
            self.comboBox_ytdlp_rate_limit.findData(settings.get_ytdlp_rate_limit())
        )
        self.groupBox_audio.setChecked(settings.get_audio())
        self.comboBox_audio_format.setCurrentIndex(
            self.comboBox_audio_format.findData(settings.get_audio_format())
        )
        self.comboBox_audio_bitrate.setCurrentIndex(
            self.comboBox_audio_bitrate.findData(settings.get_audio_bitrate())
        )
        self.comboBox_audio_normalization.setCurrentIndex(
            self.comboBox_audio_normalization.findData(
                settings.get_audio_normalization()
            )
        )
        self.checkBox_audio_embed_artwork.setChecked(settings.get_audio_embed_artwork())
        self.groupBox_video.setChecked(settings.get_video())
        self.comboBox_videocontainer.setCurrentIndex(
            self.comboBox_videocontainer.findData(settings.get_video_format())
        )
        self.comboBox_videoencoder.setCurrentIndex(
            self.comboBox_videoencoder.findData(settings.get_video_format_new())
        )
        self.groupBox_reencode_video.setChecked(settings.get_video_reencode())
        self.comboBox_videoresolution.setCurrentIndex(
            self.comboBox_videoresolution.findData(settings.get_video_resolution())
        )
        self.comboBox_fps.setCurrentIndex(
            self.comboBox_fps.findData(settings.get_video_fps())
        )
        self.checkBox_video_embed_artwork.setChecked(settings.get_video_embed_artwork())
        self.groupBox_background.setChecked(settings.get_background())
        self.checkBox_background_always.setChecked(settings.get_background_always())
        if (path := settings.get_app_path(settings.SupportedApps.KAREDI)) is not None:
            self.lineEdit_path_karedi.setText(str(path))
        if (
            path := settings.get_app_path(settings.SupportedApps.PERFORMOUS)
        ) is not None:
            self.lineEdit_path_performous.setText(str(path))
        if (
            path := settings.get_app_path(settings.SupportedApps.ULTRASTAR_MANAGER)
        ) is not None:
            self.lineEdit_path_ultrastar_manager.setText(str(path))
        if (path := settings.get_app_path(settings.SupportedApps.USDX)) is not None:
            self.lineEdit_path_usdx.setText(str(path))
        if (path := settings.get_app_path(settings.SupportedApps.VOCALUXE)) is not None:
            self.lineEdit_path_vocaluxe.setText(str(path))
        if (
            path := settings.get_app_path(settings.SupportedApps.YASS_RELOADED)
        ) is not None:
            self.lineEdit_path_yass_reloaded.setText(str(path))
        self.checkBox_discord_allowed.setChecked(settings.get_discord_allowed())

    def _setup_path_template(self) -> None:
        self.edit_path_template.textChanged.connect(self._on_path_template_changed)
        self.edit_path_template.setText(str(settings.get_path_template()))
        self.edit_path_template.setPlaceholderText(PathTemplate.default_str)
        self.button_default_path_template.pressed.connect(self.edit_path_template.clear)
        self.button_insert_placeholder.pressed.connect(
            lambda: self.edit_path_template.insert(
                self.comboBox_placeholder.currentText()
            )
        )
        for item in path_template.PathTemplatePlaceholder:
            self.comboBox_placeholder.addItem(str(item), item)
        for key in path_template.PathTemplateCustomPlaceholder.options():
            self.comboBox_placeholder.addItem(str(key))

    def _on_path_template_changed(self, text: str) -> None:
        try:
            self._path_template = (
                PathTemplate.parse(text) if text else PathTemplate.default()
            )
        except path_template.PathTemplateError as error:
            result = str(error)
            self._path_template = None
        else:
            result = str(self._path_template.evaluate(self._song).with_suffix(".txt"))
        self.edit_path_template_result.setText(result)

    def apply(self) -> None:
        self._save_settings()
        if self._browser != self.comboBox_browser.currentData():
            SessionManager.reset_session()

    def accept(self) -> None:
        if not self._save_settings():
            return
        if self._browser != self.comboBox_browser.currentData():
            SessionManager.reset_session()
        super().accept()

    def _save_settings(self) -> bool:
        new_theme = self.comboBox_theme.currentData()
        new_primary_color = self.comboBox_primary_color.currentData()
        colored_background = self.checkBox_colored_background.isChecked()
        settings.set_theme(new_theme)
        settings.set_primary_color(new_primary_color)
        settings.set_colored_background(colored_background)
        theme.apply_theme(new_theme, new_primary_color, colored_background)
        settings.set_browser(self.comboBox_browser.currentData())
        settings.set_cover(self.groupBox_cover.isChecked())
        settings.set_cover_max_size(self.comboBox_cover_max_size.currentData())
        settings.set_txt(self.groupBox_songfile.isChecked())
        settings.set_encoding(self.comboBox_encoding.currentData())
        settings.set_newline(self.comboBox_line_endings.currentData())
        settings.set_version(self.comboBox_format_version.currentData())
        settings.set_fix_linebreaks(self.comboBox_fix_linebreaks.currentData())
        settings.set_fix_first_words_capitalization(
            self.checkBox_fix_first_words_capitalization.isChecked()
        )
        settings.set_fix_spaces(self.comboBox_fix_spaces.currentData())
        settings.set_fix_quotation_marks(self.checkBox_fix_quotation_marks.isChecked())
        settings.set_throttling_threads(self.spinBox_throttling_threads.value())
        settings.set_ytdlp_rate_limit(self.comboBox_ytdlp_rate_limit.currentData())
        settings.set_audio(self.groupBox_audio.isChecked())
        settings.set_audio_format(self.comboBox_audio_format.currentData())
        settings.set_audio_bitrate(self.comboBox_audio_bitrate.currentData())
        settings.set_audio_normalization(
            self.comboBox_audio_normalization.currentData()
        )
        settings.set_audio_embed_artwork(self.checkBox_audio_embed_artwork.isChecked())
        settings.set_video(self.groupBox_video.isChecked())
        settings.set_video_format(self.comboBox_videocontainer.currentData())
        settings.set_video_format_new(self.comboBox_videoencoder.currentData())
        settings.set_video_reencode(self.groupBox_reencode_video.isChecked())
        settings.set_video_resolution(self.comboBox_videoresolution.currentData())
        settings.set_video_fps(self.comboBox_fps.currentData())
        settings.set_video_embed_artwork(self.checkBox_video_embed_artwork.isChecked())
        settings.set_background(self.groupBox_background.isChecked())
        settings.set_background_always(self.checkBox_background_always.isChecked())
        settings.set_discord_allowed(self.checkBox_discord_allowed.isChecked())
        if self._path_template:
            settings.set_path_template(self._path_template)
        else:
            QtWidgets.QMessageBox.warning(
                self, "Invalid setting", "Please provide a valid path template!"
            )
            return False
        settings.set_app_path(
            settings.SupportedApps.KAREDI, self.lineEdit_path_karedi.text()
        )
        settings.set_app_path(
            settings.SupportedApps.PERFORMOUS, self.lineEdit_path_performous.text()
        )
        settings.set_app_path(
            settings.SupportedApps.ULTRASTAR_MANAGER,
            self.lineEdit_path_ultrastar_manager.text(),
        )
        settings.set_app_path(
            settings.SupportedApps.USDX, self.lineEdit_path_usdx.text()
        )
        settings.set_app_path(
            settings.SupportedApps.VOCALUXE, self.lineEdit_path_vocaluxe.text()
        )
        settings.set_app_path(
            settings.SupportedApps.YASS_RELOADED,
            self.lineEdit_path_yass_reloaded.text(),
        )
        return True



"""Functions for Disord integration."""

import requests

from usdb_syncer import SongId, db, remote_config
from usdb_syncer.constants import Usdb
from usdb_syncer.logger import Logger
from usdb_syncer.usdb_song import UsdbSong

DISCORD_COLOR_RED = 0xED4245  # equivalent to discord.Color.red()


def notify_discord(
    song_id: SongId, url: str, kind: str, error_str: str, logger: Logger
) -> None:
    """Notify unavailable resources on Discord (without using the discord package)."""
    if not (song := UsdbSong.get(song_id)):
        logger.debug("Song id does not exist.")
        return

    if not (dwh_url := remote_config.discord_webhook_url()):
        logger.debug("No discord webhook configured. Skipping Discord notification.")
        return

    with db.transaction():
        if not db.maybe_insert_discord_notification(song_id, url):
            logger.debug(
                "Failed resource was already reported. Skipping Discord notification."
            )
            return

        embed = {
            "color": DISCORD_COLOR_RED,
            "author": {
                "name": f"{song_id}: {song.artist} - {song.title}",
                "url": song_id.usdb_detail_url(),
                "icon_url": f"{Usdb.COVER_URL}{song_id:d}.jpg",
            },
            "fields": [{"name": f"{kind} {error_str}:", "value": url, "inline": False}],
        }

        payload = {"embeds": [embed]}

        response = requests.post(dwh_url, json=payload, timeout=5)
        response.raise_for_status()



"""Functionality related to the usdb.animux.de web page."""

import re
import time
from collections.abc import Iterator
from datetime import datetime
from enum import Enum
from typing import Any, assert_never

import attrs
import requests
from bs4 import BeautifulSoup, NavigableString, Tag
from requests import Session

from usdb_syncer import SongId, errors, settings
from usdb_syncer.constants import (
    SUPPORTED_VIDEO_SOURCES_REGEX,
    Usdb,
    UsdbStrings,
    UsdbStringsEnglish,
    UsdbStringsFrench,
    UsdbStringsGerman,
)
from usdb_syncer.logger import Logger, logger, song_logger
from usdb_syncer.usdb_song import UsdbSong
from usdb_syncer.utils import extract_youtube_id, normalize

SONG_LIST_ROW_REGEX = re.compile(
    r'<td(?:.*?<source src="(?P<sample_url>.*?)".*?)?></td>'
    r'<td onclick="show_detail\((?P<song_id>\d+)\)".*?>'
    r'<img src="(?P<cover_url>.*?)".*?></td>'
    r'<td onclick="show_detail\(\d+\)">(?P<artist>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)"><a href=.*?>(?P<title>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<genre>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<year>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<edition>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<golden_notes>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<language>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<creator>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<rating>.*?)</td>\n'
    r'<td onclick="show_detail\(\d+\)">(?P<views>.*?)</td>'
)
WELCOME_REGEX = re.compile(
    r"<td class='row3' colspan='2'>\s*<span class='gen'>([^<]+) <b>([^<]+)</b>"
)


def establish_usdb_login(session: Session) -> bool:
    """Tries to log in to USDB if necessary. Returns final login status."""
    if user := get_logged_in_usdb_user(session):
        logger.info(f"Using existing login of USDB user '{user}'.")
        return True
    if (auth := settings.get_usdb_auth())[0] and auth[1]:
        if login_to_usdb(session, *auth):
            logger.info(f"Successfully logged in to USDB with user '{auth[0]}'.")
            return True
        logger.error(f"Login to USDB with user '{auth[0]}' failed!")
    else:
        logger.warning(
            "Not logged in to USDB. Please go to 'Synchronize > USDB Login', then "
            "select the browser you are logged in with and/or enter your credentials."
        )
    return False


def new_session_with_cookies(browser: settings.Browser) -> Session:
    session = Session()
    if cookies := browser.cookies():
        for cookie in cookies:
            session.cookies.set_cookie(cookie)
    return session


class SessionManager:
    """Singleton for managing the global session instance."""

    _session: Session | None = None
    _connecting: bool = False

    @classmethod
    def session(cls) -> Session:
        while cls._connecting:
            time.sleep(0.1)
        if cls._session is None:
            cls._connecting = True
            try:
                cls._session = new_session_with_cookies(settings.get_browser())
                establish_usdb_login(cls._session)
            finally:
                cls._connecting = False
        return cls._session

    @classmethod
    def reset_session(cls) -> None:
        if cls._session:
            cls._session.close()
            cls._session = None

    @classmethod
    def has_session(cls) -> bool:
        return cls._session is not None


def get_logged_in_usdb_user(session: Session) -> str | None:
    response = session.get(Usdb.BASE_URL, timeout=10, params={"link": "profil"})
    response.raise_for_status()
    if match := WELCOME_REGEX.search(response.text):
        return match.group(2)
    return None


def login_to_usdb(session: Session, user: str, password: str) -> bool:
    """True if success."""
    response = session.post(
        Usdb.BASE_URL,
        timeout=10,
        data={"user": user, "pass": password, "login": "Login"},
    )
    response.raise_for_status()
    return UsdbStrings.LOGIN_INVALID not in response.text


def log_out_of_usdb(session: Session) -> None:
    session.post(Usdb.BASE_URL, timeout=10, params={"link": "logout"})


class RequestMethod(Enum):
    """Supported HTTP requests."""

    GET = "GET"
    POST = "POST"


@attrs.define(kw_only=True)
class CommentContents:
    """The parsed contents of a SongComment."""

    text: str
    youtube_ids: list[str]
    urls: list[str]


class SongComment:
    """A comment to a song on USDB."""

    date_time: datetime
    author: str
    contents: CommentContents

    def __init__(
        self, *, date_time: str, author: str, contents: CommentContents
    ) -> None:
        self.date_time = datetime.strptime(date_time, Usdb.DATETIME_STRF)
        self.author = author
        self.contents = contents


@attrs.define
class SongDetails:
    """Details about a song that USDB shows on a song's page, or are specified in the
    comment section."""

    song_id: SongId
    artist: str
    title: str
    cover_url: str | None
    language: str
    year: int | None
    genre: str
    edition: str
    bpm: float
    gap: float
    golden_notes: bool
    song_check: bool
    date_time: datetime
    uploader: str
    editors: list[str]
    views: int
    rating: int
    votes: int
    audio_sample: str | None
    comments: list[SongComment] = attrs.field(factory=list)

    def all_comment_videos(self) -> Iterator[str]:
        """Yields all parsed URLs and YouTube ids. Order is latest to earliest, then ids
        before URLs.
        """
        for comment in self.comments:
            yield from comment.contents.youtube_ids
            yield from comment.contents.urls


def get_usdb_page(
    rel_url: str,
    method: RequestMethod = RequestMethod.GET,
    *,
    headers: dict[str, str] | None = None,
    payload: dict[str, str] | None = None,
    params: dict[str, str] | None = None,
    session: Session | None = None,
) -> str:
    """Retrieve HTML subpage from USDB.

    Parameters:
        rel_url: relative url of page to retrieve
        method: GET or POST
        headers: dict of headers to send with request
        payload: dict of data to send with request
        params: dict of params to send with request
        session: Session to use instead of the global one
    """
    existing_session = SessionManager.has_session()

    def page() -> str:
        return _get_usdb_page_inner(
            session or SessionManager.session(),
            rel_url,
            method=method,
            headers=headers,
            payload=payload,
            params=params,
        )

    try:
        return page()
    except requests.ConnectionError:
        logger.debug("Connection failed; session may have expired; retrying ...")
    except errors.UsdbLoginError:
        # skip login retry if custom or just created session
        if session or not existing_session:
            raise
        logger.debug(f"Page '{rel_url}' is private; trying to log in ...")
    if not session:
        SessionManager.reset_session()
    return page()


def _get_usdb_page_inner(
    session: Session,
    rel_url: str,
    method: RequestMethod = RequestMethod.GET,
    *,
    headers: dict[str, str] | None = None,
    payload: dict[str, str] | None = None,
    params: dict[str, str] | None = None,
) -> str:
    session = session or SessionManager.session()
    url = Usdb.BASE_URL + rel_url
    match method:
        case RequestMethod.GET:
            logger.debug(f"Get request for {url}")
            response = session.get(url, headers=headers, params=params, timeout=10)
        case RequestMethod.POST:
            logger.debug(f"Post request for {url}")
            response = session.post(
                url, headers=headers, data=payload, params=params, timeout=10
            )
        case _ as unreachable:
            assert_never(unreachable)
    response.raise_for_status()
    response.encoding = "utf-8"
    if UsdbStrings.NOT_LOGGED_IN in (page := normalize(response.text)):
        raise errors.UsdbLoginError
    if UsdbStrings.DATASET_NOT_FOUND in page:
        raise errors.UsdbNotFoundError
    return page


def get_usdb_details(song_id: SongId) -> SongDetails:
    """Retrieve song details from usdb webpage, if song exists.

    Parameters:
        song_id: id of song to retrieve details for
    """
    html = get_usdb_page(
        "index.php", params={"id": str(int(song_id)), "link": "detail"}
    )
    return _parse_song_page(BeautifulSoup(html, "lxml"), song_id)


def _parse_song_page(soup: BeautifulSoup, song_id: SongId) -> SongDetails:
    logger = song_logger(song_id)
    usdb_strings = _usdb_strings_from_soup(soup)
    details_table, comments_table, *_ = soup.find_all("table", border="0", width="500")
    assert isinstance(details_table, Tag)
    assert isinstance(comments_table, Tag)
    details = _parse_details_table(details_table, song_id, usdb_strings, logger)
    details.comments = _parse_comments_table(comments_table, logger)
    return details


def _usdb_strings_from_soup(soup: BeautifulSoup) -> type[UsdbStrings]:
    span = soup.find("span", class_="gen")
    assert span
    return _usdb_strings_from_welcome(span.text.split(" ", 1)[0].removesuffix(","))


def _usdb_strings_from_html(html: str) -> type[UsdbStrings]:
    if match := WELCOME_REGEX.search(html):
        return _usdb_strings_from_welcome(match.group(1))
    raise errors.UsdbUnknownLanguageError


def _usdb_strings_from_welcome(welcome_string: str) -> type[UsdbStrings]:
    match welcome_string:
        case UsdbStringsEnglish.WELCOME:
            return UsdbStringsEnglish
        case UsdbStringsGerman.WELCOME:
            return UsdbStringsGerman
        case UsdbStringsFrench.WELCOME:
            return UsdbStringsFrench
    raise errors.UsdbUnknownLanguageError


def get_usdb_available_songs(
    max_skip_id: SongId,
    content_filter: dict[str, str] | None = None,
    session: Session | None = None,
) -> list[UsdbSong]:
    """Return a list of all available songs.

    Parameters:
        max_skip_id: only fetch ids larger than this
        content_filter: filters response (e.g. {'artist': 'The Beatles'})
    """
    available_songs: list[UsdbSong] = []
    payload = {
        "order": "id",
        "ud": "desc",
        "limit": str(Usdb.MAX_SONGS_PER_PAGE),
        "details": "1",
    }
    payload.update(content_filter or {})
    for start in range(0, Usdb.MAX_SONG_ID, Usdb.MAX_SONGS_PER_PAGE):
        payload["start"] = str(start)
        html = get_usdb_page(
            "index.php",
            RequestMethod.POST,
            params={"link": "list"},
            payload=payload,
            session=session,
        )
        songs = [
            song
            for song in _parse_songs_from_songlist(html)
            if song.song_id > max_skip_id
        ]
        available_songs.extend(songs)

        if len(songs) < Usdb.MAX_SONGS_PER_PAGE:
            break

    logger.info(f"Fetched {len(available_songs)} new song(s) from USDB.")
    return available_songs


def _parse_songs_from_songlist(html: str) -> Iterator[UsdbSong]:
    return (
        UsdbSong.from_html(
            _usdb_strings_from_html(html),
            sample_url=match["sample_url"] or "",
            song_id=match["song_id"],
            artist=match["artist"],
            title=match["title"],
            genre=match["genre"],
            year=match["year"],
            edition=match["edition"],
            golden_notes=match["golden_notes"],
            language=match["language"],
            creator=match["creator"],
            rating=match["rating"],
            views=match["views"],
        )
        for match in SONG_LIST_ROW_REGEX.finditer(html)
    )


def _parse_details_table(
    details_table: Tag, song_id: SongId, usdb_strings: type[UsdbStrings], logger: Logger
) -> SongDetails:
    """Parse song attributes from usdb page.

    Parameters:
        details: dict of song attributes
        details_table: BeautifulSoup object of song details table
    """
    editors = []
    pointer = details_table.find(string=usdb_strings.SONG_EDITED_BY)
    while pointer is not None:
        pointer = pointer.find_next("td")
        if pointer.a is None:  # type: ignore
            break
        editors.append(pointer.text.strip())  # type: ignore
        pointer = pointer.find_next("tr")  # type: ignore

    stars = details_table.find(string=usdb_strings.SONG_RATING).next.find_all("img")  # type: ignore
    votes_str = details_table.find(string=usdb_strings.SONG_RATING).next_element.text  # type: ignore

    audio_sample = ""
    if param := details_table.find("source"):
        assert isinstance(param, Tag)
        src = param.get("src")
        assert isinstance(src, str)
        audio_sample = src
    else:
        logger.debug("No audio sample found. Consider adding one!")

    cover_url = details_table.img["src"]  # type: ignore
    assert isinstance(cover_url, str)
    if "nocover" in cover_url:
        logger.debug("No USDB cover. Consider adding one!")

    year_str = _find_text_after(details_table, usdb_strings.SONG_YEAR)
    year = int(year_str) if len(year_str) == 4 and year_str.isdigit() else None

    return SongDetails(
        song_id=song_id,
        artist=details_table.find_next("td").text,  # type: ignore
        title=details_table.find_next("td").find_next("td").text,  # type: ignore
        cover_url=None if "nocover" in cover_url else Usdb.BASE_URL + cover_url,
        language=_find_text_after(details_table, usdb_strings.SONG_LANGUAGE),
        year=year,
        genre=_find_text_after(details_table, "Genre"),
        edition=_find_text_after(details_table, "Edition"),
        bpm=float(_find_text_after(details_table, "BPM").replace(",", ".")),
        gap=float(_find_text_after(details_table, "GAP").replace(",", ".") or 0),
        golden_notes=_find_text_after(details_table, usdb_strings.GOLDEN_NOTES)
        == usdb_strings.YES,
        song_check=_find_text_after(details_table, usdb_strings.SONGCHECK)
        == usdb_strings.YES,
        date_time=datetime.strptime(
            _find_text_after(details_table, usdb_strings.DATE), Usdb.DATETIME_STRF
        ),
        uploader=_find_text_after(details_table, usdb_strings.UPLOADED_BY),
        editors=editors,
        views=int(_find_text_after(details_table, usdb_strings.VIEWS)),
        rating=sum("star.png" in s.get("src") for s in stars),
        votes=int(votes_str.split("(")[1].split(")")[0]),
        audio_sample=audio_sample or None,
    )


def _find_text_after(details_table: Tag, label: str) -> str:
    if isinstance((tag := details_table.find(string=label)), NavigableString):
        if isinstance(tag.next, Tag):
            return tag.next.text.strip()
    raise errors.UsdbParseError(f"Text after {label} not found.")  # noqa: TRY003


def _parse_comments_table(comments_table: Tag, logger: Logger) -> list[SongComment]:
    """Parse the table into individual comments, extracting potential video links,
    GAP and BPM values.
    """
    comments = []
    # last entry is the field to enter a new comment, so this one is ignored
    for header in comments_table.find_all("tr", class_="list_tr2")[:-1]:
        assert isinstance(header, Tag)
        td = header.find("td")
        assert isinstance(td, Tag)
        meta = td.text.strip()
        if " | " not in meta:
            # header is just the placeholder element
            break
        date_time, author = meta.removeprefix("[del] [edit] ").split(" | ")
        assert isinstance(header.next_sibling, Tag)
        contents = _parse_comment_contents(header.next_sibling, logger)
        comments.append(
            SongComment(date_time=date_time, author=author, contents=contents)
        )

    return comments


def _parse_comment_contents(contents: Tag, logger: Logger) -> CommentContents:
    td_element = contents.find("td")
    assert isinstance(td_element, Tag)
    for emoji in td_element.find_all("img"):
        assert isinstance(emoji, Tag)
        title = emoji.get("title")
        assert isinstance(title, str)
        emoji.replace_with(NavigableString(title))

    text = td_element.text.strip()  # type: ignore
    urls: list[str] = []
    youtube_ids: list[str] = []

    for url in _all_urls_in_comment(contents, text, logger):
        if yt_id := extract_youtube_id(url):
            youtube_ids.append(yt_id)
        else:
            urls.append(url)

    return CommentContents(text=text, urls=urls, youtube_ids=youtube_ids)


def _all_urls_in_comment(contents: Tag, text: str, logger: Logger) -> Iterator[str]:
    for embed in contents.find_all("embed"):
        if src := _extract_supported_video_source(embed, "src"):
            logger.debug("Video embed found. Consider embedding as iframe.")
            yield src
    for iframe in contents.find_all("iframe"):
        if src := _extract_supported_video_source(iframe, "src"):
            yield src
    for anchor in contents.find_all("a"):
        if url := _extract_supported_video_source(anchor, "href"):
            logger.debug("Video href found. Consider embedding as iframe.")
            yield url
    for match in SUPPORTED_VIDEO_SOURCES_REGEX.finditer(text):
        logger.debug("Video plain url found. Consider embedding as iframe.")
        yield match.group(1)


def _extract_supported_video_source(tag: Any, attr_key: str) -> str | None:
    if (
        isinstance(tag, Tag)
        and isinstance(src := tag.get(attr_key), str)
        and SUPPORTED_VIDEO_SOURCES_REGEX.fullmatch(src)
    ):
        return src
    return None


def get_notes(song_id: SongId, logger: Logger) -> str:
    """Retrieve notes for a song."""
    logger.debug("fetching notes")
    html = get_usdb_page(
        "index.php",
        RequestMethod.POST,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        params={"link": "gettxt", "id": str(int(song_id))},
        payload={"wd": "1"},
    )
    return _parse_song_txt_from_txt_page(BeautifulSoup(html, "lxml"))


def _parse_song_txt_from_txt_page(soup: BeautifulSoup) -> str:
    if isinstance(textarea := soup.find("textarea"), Tag):
        return textarea.string or ""
    raise errors.UsdbParseError("textarea for notes not found")  # noqa: TRY003


def post_song_comment(song_id: SongId, text: str, rating: str) -> None:
    """Post a song comment to USDB."""
    payload = {"text": text, "stars": rating}

    get_usdb_page(
        "index.php",
        RequestMethod.POST,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        params={"link": "detail", "id": str(int(song_id)), "comment": str(1)},
        payload=payload,
    )
    logger = song_logger(song_id)
    logger.debug("Comment posted on USDB.")


def post_song_rating(song_id: SongId, stars: int) -> None:
    """Post a song rating to USDB."""

    payload = {"stars": str(stars), "text": "onlyvoting"}

    get_usdb_page(
        "index.php",
        RequestMethod.POST,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        params={"link": "detail", "id": str(int(song_id)), "comment": str(1)},
        payload=payload,
    )
    logger = song_logger(song_id)
    logger.debug(f"{stars}-star rating posted on USDB.")



"""Contains a runnable song loader."""

from __future__ import annotations

import copy
import shutil
import tempfile
import time
from collections.abc import Iterable, Iterator
from itertools import islice
from pathlib import Path
from typing import ClassVar, assert_never

import attrs
import send2trash
import shiboken6
from PySide6 import QtCore

from usdb_syncer import (
    SongId,
    SyncMetaId,
    db,
    download_options,
    errors,
    events,
    hooks,
    resource_dl,
    settings,
    usdb_scraper,
    utils,
)
from usdb_syncer.custom_data import CustomData
from usdb_syncer.discord import notify_discord
from usdb_syncer.logger import Logger, logger, song_logger
from usdb_syncer.postprocessing import write_audio_tags, write_video_tags
from usdb_syncer.resource_dl import ResourceDLError
from usdb_syncer.settings import FormatVersion
from usdb_syncer.song_txt import SongTxt
from usdb_syncer.sync_meta import ResourceFile, SyncMeta
from usdb_syncer.usdb_scraper import SongDetails
from usdb_syncer.usdb_song import DownloadStatus, UsdbSong
from usdb_syncer.utils import video_url_from_resource


class DownloadManager:
    """Manager for concurrent song downloads."""

    _jobs: ClassVar[dict[SongId, _SongLoader]] = {}
    _pause = False
    _pool: QtCore.QThreadPool | None = None

    @classmethod
    def download(cls, songs: Iterable[UsdbSong]) -> None:
        options = download_options.download_options()
        for song in songs:
            if song.song_id in cls._jobs:
                cls._jobs[song.song_id].logger.warning("Already downloading!")
                continue
            cls._jobs[song.song_id] = job = _SongLoader(song, options)
            job.pause = cls._pause
            cls._threadpool().start(job)

    @classmethod
    def abort(cls, songs: Iterable[SongId]) -> None:
        for song in songs:
            if (job := cls._jobs.get(song)) and shiboken6.isValid(job):
                if cls._threadpool().tryTake(job):
                    job.logger.info("Download aborted by user request.")
                    job.song.status = DownloadStatus.NONE
                    with db.transaction():
                        job.song.upsert()
                    events.SongChanged(job.song_id).post()
                    events.DownloadFinished(job.song_id).post()
                else:
                    job.abort = True

    @classmethod
    def set_pause(cls, pause: bool) -> None:
        cls._pause = pause
        for job in cls._jobs.values():
            job.pause = pause

    @classmethod
    def quit(cls) -> None:
        if cls._pool:
            logger.debug(f"Quitting {len(cls._jobs)} downloads.")
            for job in cls._jobs.values():
                job.abort = True
            cls._pool.waitForDone()

    @classmethod
    def _threadpool(cls) -> QtCore.QThreadPool:
        if cls._pool is None:
            cls._pool = QtCore.QThreadPool()
            if threads := settings.get_throttling_threads():
                cls._pool.setMaxThreadCount(threads)
            events.DownloadFinished.subscribe(cls._remove_job)
        return cls._pool

    @classmethod
    def _remove_job(cls, event: events.DownloadFinished) -> None:
        if event.song_id in cls._jobs:
            del cls._jobs[event.song_id]


@attrs.define(kw_only=True)
class _Locations:
    """Paths for downloading a song."""

    _current: Path | None
    # includes filename stem
    _target: Path
    _tempdir: Path

    @classmethod
    def new(
        cls, song: UsdbSong, options: download_options.Options, tempdir: Path
    ) -> _Locations:
        target = options.path_template.evaluate(song, options.song_dir)
        if (
            _current := song.sync_meta.path.parent if song.sync_meta else None
        ) and utils.path_matches_maybe_with_suffix(_current, target.parent):
            target = _current / target.name
        else:
            target = utils.next_unique_directory(target.parent) / target.name
        return cls(current=_current, target=target, tempdir=tempdir)  # pyright: ignore

    def current_path(self, file: str = "", ext: str = "") -> Path | None:
        """Path to file in the current download directory if it exists.
        The final path component is the generic name or the provided file, optionally
        with the provided extension joined with a '.' unless one is already present.
        """
        return self._path(self._current, file, ext) if self._current else None

    def temp_path(self, file: str = "", ext: str = "") -> Path:
        """Path to file in the temporary download directory.
        The final path component is the generic name or the provided file, optionally
        with the provided extension joined with a '.' unless one is already present.
        """
        return self._path(self._tempdir, file, ext)

    def target_path(self, file: str = "", ext: str = "") -> Path:
        """Path to file in the final download directory.
        The final path component is the generic name or the provided file, optionally
        with the provided extension joined with a '.' unless one is already present.
        """
        return self._path(self._target.parent, file, ext)

    def _path(self, parent: Path, file: str = "", ext: str = "") -> Path:
        name = file or self._target.name
        if ext:
            name = f"{name}{'' if '.' in ext else '.'}{ext}"
        return parent.joinpath(name)

    def filename(self, ext: str = "") -> str:
        if ext:
            return f"{self._target.name}{'' if '.' in ext else '.'}{ext}"
        return self._target.name

    def move_to_target_folder(self) -> None:
        """Rename the path of the song folder if it does not match the template, and
        ensure the target directory exists.
        """
        if self._current and self._current != self._target:
            self._target.parent.parent.mkdir(parents=True, exist_ok=True)
            self._current.rename(self._target.parent)
            self._current = self._target.parent
        else:
            self._target.parent.mkdir(parents=True, exist_ok=True)


@attrs.define
class _TempResourceFile:
    """Interim resource file in the temporary folder, or in the old folder if the
    resource is potentially kept.
    """

    old_fname: str | None = None
    new_fname: str | None = None
    resource: str | None = None

    def path_and_resource(
        self, locations: _Locations, temp: bool
    ) -> tuple[Path, str] | None:
        if (path := self.path(locations, temp=temp)) and self.resource:
            return (path, self.resource)
        return None

    def path(self, locations: _Locations, temp: bool) -> Path | None:
        if self.new_fname:
            if temp:
                return locations.temp_path(self.new_fname)
            return locations.target_path(self.new_fname)
        if self.old_fname:
            return locations.current_path(self.old_fname)
        return None

    def to_resource_file(
        self, locations: _Locations, temp: bool
    ) -> ResourceFile | None:
        if path_resource := self.path_and_resource(locations, temp=temp):
            return ResourceFile.new(*path_resource)
        return None


@attrs.define
class _TempResourceFiles:
    """Collection of all temporary resource files."""

    txt: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    audio: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    video: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    cover: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    background: _TempResourceFile = attrs.field(factory=_TempResourceFile)

    def __iter__(self) -> Iterator[_TempResourceFile]:
        return iter((self.txt, self.audio, self.video, self.cover, self.background))


@attrs.define
class _Context:
    """_Context for downloading media and creating a song folder."""

    # deep copy of the passed in song
    song: UsdbSong
    details: SongDetails
    options: download_options.Options
    txt: SongTxt
    locations: _Locations
    logger: Logger
    out: _TempResourceFiles = attrs.field(factory=_TempResourceFiles)

    def __attrs_post_init__(self) -> None:
        # reuse old resource files unless we acquire new ones later on
        # txt is always rewritten
        if self.song.sync_meta and (current := self.locations.current_path()):
            for old, out in (
                (self.song.sync_meta.audio, self.out.audio),
                (self.song.sync_meta.video, self.out.video),
                (self.song.sync_meta.cover, self.out.cover),
                (self.song.sync_meta.background, self.out.background),
            ):
                if old and old.is_in_sync(current.parent):
                    out.resource = old.resource
                    out.old_fname = old.fname

    @classmethod
    def new(
        cls,
        song: UsdbSong,
        options: download_options.Options,
        tempdir: Path,
        log: Logger,
    ) -> _Context:
        song = copy.deepcopy(song)
        details, txt = _get_usdb_data(song.song_id, options.txt_options, log)
        _update_song_with_usdb_data(song, details, txt)
        paths = _Locations.new(song, options, tempdir)
        if not song.sync_meta:
            song.sync_meta = SyncMeta.new(
                song.song_id, paths.target_path().parent, txt.meta_tags
            )
        return cls(song, details, options, txt, paths, log)

    def all_audio_resources(self) -> Iterator[str]:
        if self.txt.meta_tags.audio:
            yield self.txt.meta_tags.audio
        if not self.txt.meta_tags.video:
            self.logger.debug("No valid audio/video meta tag. Looking in comments.")
        yield from self.all_video_resources()

    def all_video_resources(self) -> Iterator[str]:
        if self.txt.meta_tags.video:
            yield self.txt.meta_tags.video
        yield from self.details.all_comment_videos()

    def background_url(self) -> str | None:
        url = None
        if self.txt.meta_tags.background:
            url = self.txt.meta_tags.background.source_url(self.logger)
            self.logger.debug(f"downloading background from #VIDEO params: {url}")
        return url


def _get_usdb_data(
    song_id: SongId, txt_options: download_options.TxtOptions | None, log: Logger
) -> tuple[SongDetails, SongTxt]:
    details = usdb_scraper.get_usdb_details(song_id)
    log.info(f"Found '{details.artist} - {details.title}' on USDB.")
    txt_str = usdb_scraper.get_notes(details.song_id, log)
    txt = SongTxt.parse(txt_str, log)
    txt.sanitize(txt_options)
    txt.headers.creator = txt.headers.creator or details.uploader or None
    return details, txt


def _update_song_with_usdb_data(
    song: UsdbSong, details: SongDetails, txt: SongTxt
) -> None:
    song.artist = details.artist
    song.title = details.title
    song.language = txt.headers.language or ""
    song.edition = txt.headers.edition or ""
    song.golden_notes = details.golden_notes
    song.rating = details.rating
    song.views = details.views
    if txt.headers.year and len(txt.headers.year) == 4 and txt.headers.year.isdigit():
        song.year = int(txt.headers.year)
    else:
        song.year = None
    song.genre = txt.headers.genre or ""
    song.creator = txt.headers.creator or ""


class _SongLoader(QtCore.QRunnable):
    """Runnable to create a complete song folder."""

    abort = False
    pause = False

    def __init__(self, song: UsdbSong, options: download_options.Options) -> None:
        super().__init__()
        self.song = song
        self.song_id = song.song_id
        self.options = options
        self.logger = song_logger(self.song_id)

    def run(self) -> None:
        with db.managed_connection(utils.AppPaths.db):
            try:
                self.song = self._run_inner()
            except errors.AbortError:
                self.logger.info("Download aborted by user request.")
                self.song.status = DownloadStatus.NONE
            except errors.UsdbLoginError:
                self.logger.error("Aborted; download requires login.")  # noqa: TRY400
                self.song.status = DownloadStatus.FAILED
            except errors.UsdbNotFoundError:
                self.logger.error("Song has been deleted from USDB.")  # noqa: TRY400
                with db.transaction():
                    self.song.delete()
                if meta := self.song.sync_meta:
                    path = meta.path.parent
                    self.logger.info(f"Trashing local song {path}")
                    send2trash.send2trash(path)
                events.SongDeleted(self.song_id).post()
                events.DownloadFinished(self.song_id).post()
                return
            except Exception:
                self.logger.exception(
                    "Failed to finish download due to an unexpected error. "
                    "See debug log for more information."
                )
                self.song.status = DownloadStatus.FAILED
            else:
                self.song.status = DownloadStatus.NONE
                self.logger.info("All done!")
            with db.transaction():
                self.song.upsert()
        events.SongChanged(self.song_id).post()
        events.DownloadFinished(self.song_id).post()

    def _run_inner(self) -> UsdbSong:
        self._check_flags()
        self.song.status = DownloadStatus.DOWNLOADING
        with db.transaction():
            self.song.upsert()
        events.SongChanged(self.song_id).post()
        with tempfile.TemporaryDirectory() as tempdir:
            ctx = _Context.new(self.song, self.options, Path(tempdir), self.logger)
            for job in (
                _maybe_download_audio,
                _maybe_download_video,
                _maybe_download_cover,
                _maybe_download_background,
                _maybe_write_audio_tags,
                _maybe_write_video_tags,
            ):
                self._check_flags()
                job(ctx)

            # last chance to abort before irreversible changes
            self._check_flags()
            _cleanup_existing_resources(ctx)
            # only here so filenames in header are up-to-date
            _maybe_write_txt(ctx)
            ctx.locations.move_to_target_folder()
            _persist_tempfiles(ctx)
        _write_sync_meta(ctx)
        hooks.SongLoaderDidFinish.call(ctx.song)
        return ctx.song

    def _check_flags(self) -> None:
        if self.abort:
            raise errors.AbortError
        if self.pause:
            while self.pause:
                time.sleep(0.5)
                if self.abort:
                    raise errors.AbortError


def _maybe_download_audio(ctx: _Context) -> None:
    if not (options := ctx.options.audio_options):
        return
    for resource in islice(ctx.all_audio_resources(), 10):
        if ctx.out.audio.resource == resource:
            ctx.logger.info("Audio resource is unchanged.")
            return
        dl_result = resource_dl.download_audio(
            resource,
            options,
            ctx.options.browser,
            ctx.locations.temp_path(),
            ctx.logger,
        )
        if ext := dl_result.extension:
            ctx.out.audio.resource = resource
            ctx.out.audio.new_fname = ctx.locations.filename(ext=ext)
            ctx.logger.info("Success! Downloaded audio.")
            return
        if dl_result.error in {
            resource_dl.ResourceDLError.RESOURCE_INVALID,
            resource_dl.ResourceDLError.RESOURCE_UNSUPPORTED,
            resource_dl.ResourceDLError.RESOURCE_UNAVAILABLE,
            resource_dl.ResourceDLError.RESOURCE_PARSE_ERROR,
        }:
            if ctx.options.notify_discord and (
                url := video_url_from_resource(resource)
            ):
                notify_discord(
                    ctx.song.song_id, url, "Audio", dl_result.error.value, logger
                )
    keep = " Keeping last resource." if ctx.out.audio.resource else ""
    song_len = ctx.txt.minimum_song_length()
    ctx.logger.error(f"Failed to download audio (song duration > {song_len})!{keep}")


def _maybe_download_video(ctx: _Context) -> None:
    if not (options := ctx.options.video_options) or ctx.txt.meta_tags.is_audio_only():
        return
    for resource in islice(ctx.all_video_resources(), 10):
        if ctx.out.video.resource == resource:
            ctx.logger.info("Video resource is unchanged.")
            return
        dl_result = resource_dl.download_video(
            resource,
            options,
            ctx.options.browser,
            ctx.locations.temp_path(),
            ctx.logger,
        )
        if ext := dl_result.extension:
            ctx.out.video.resource = resource
            ctx.out.video.new_fname = ctx.locations.filename(ext=ext)
            ctx.logger.info("Success! Downloaded video.")
            return
        if dl_result.error in {
            resource_dl.ResourceDLError.RESOURCE_INVALID,
            resource_dl.ResourceDLError.RESOURCE_UNSUPPORTED,
            resource_dl.ResourceDLError.RESOURCE_UNAVAILABLE,
            resource_dl.ResourceDLError.RESOURCE_PARSE_ERROR,
        }:
            if ctx.options.notify_discord and (
                url := video_url_from_resource(resource)
            ):
                notify_discord(
                    ctx.song.song_id, url, "Video", dl_result.error.value, logger
                )
    keep = " Keeping last resource." if ctx.out.video.resource else ""
    ctx.logger.error(f"Failed to download video!{keep}")


def _maybe_download_cover(ctx: _Context) -> None:
    if not ctx.options.cover:
        return
    if ctx.txt.meta_tags.cover is None and ctx.details.cover_url is None:
        ctx.logger.warning("No cover resource found.")
        return
    if cover := ctx.txt.meta_tags.cover:
        url = cover.source_url(ctx.logger)
        if _download_cover_url(ctx, url):
            return
        if ctx.options.notify_discord:
            notify_discord(
                ctx.song.song_id,
                url,
                "Cover",
                ResourceDLError.RESOURCE_UNAVAILABLE.value,
                ctx.logger,
            )
    if ctx.details.cover_url:
        ctx.logger.warning("Falling back to small USDB cover.")
        if _download_cover_url(ctx, ctx.details.cover_url, process=False):
            return
    keep = " Keeping last resource." if ctx.out.cover.resource else ""
    ctx.logger.error(f"Failed to download cover!{keep}")


def _download_cover_url(ctx: _Context, url: str, process: bool = True) -> bool:
    """True if download was successful (or is unnecessary)."""
    assert ctx.options.cover
    if ctx.out.cover.resource == url:
        if sync_meta := ctx.song.sync_meta:
            if sync_meta.meta_tags.cover == ctx.txt.meta_tags.cover:
                ctx.logger.info(
                    "Cover resource and postprocessing parameters are unchanged, "
                    "skipping."
                )
                return True
            ctx.logger.info(
                "Cover postprocessing parameters have changed, redownloading."
            )
    if path := resource_dl.download_and_process_image(
        url=url,
        target_stem=ctx.locations.temp_path(),
        meta_tags=ctx.txt.meta_tags.cover,
        details=ctx.details,
        kind=resource_dl.ImageKind.COVER,
        max_width=ctx.options.cover.max_size,
        process=process,
    ):
        ctx.out.cover.resource = url
        ctx.out.cover.new_fname = path.name
        ctx.logger.info("Success! Downloaded cover.")
        return True
    return False


def _maybe_download_background(ctx: _Context) -> None:
    if not (options := ctx.options.background_options):
        return
    if not options.download_background(bool(ctx.out.video.resource)):
        return
    if not (url := ctx.background_url()):
        ctx.logger.warning("No background resource found.")
        return
    if ctx.out.background.resource == url:
        if sync_meta := ctx.song.sync_meta:
            if sync_meta.meta_tags.background == ctx.txt.meta_tags.background:
                ctx.logger.info(
                    "Background resource and postprocessing parameters are unchanged, "
                    "skipping."
                )
                return
            ctx.logger.info(
                "Background postprocessing parameters have changed, redownloading."
            )
    if path := resource_dl.download_and_process_image(
        url=url,
        target_stem=ctx.locations.temp_path(),
        meta_tags=ctx.txt.meta_tags.background,
        details=ctx.details,
        kind=resource_dl.ImageKind.BACKGROUND,
        max_width=None,
    ):
        ctx.out.background.resource = url
        ctx.out.background.new_fname = path.name
        ctx.logger.info("Success! Downloaded background.")
    else:
        if ctx.options.notify_discord:
            notify_discord(
                ctx.song.song_id,
                url,
                "Background",
                ResourceDLError.RESOURCE_UNAVAILABLE.value,
                ctx.logger,
            )
        keep = " Keeping last resource." if ctx.out.cover.resource else ""
        ctx.logger.error(f"Failed to download background!{keep}")


def _maybe_write_txt(ctx: _Context) -> None:
    if not (options := ctx.options.txt_options):
        return
    _write_headers(ctx)
    path = ctx.locations.temp_path(ext="txt")
    ctx.out.txt.new_fname = path.name
    ctx.txt.write_to_file(path, options.encoding.value, options.newline.value)
    ctx.out.txt.resource = ctx.song.song_id.usdb_gettxt_url()
    ctx.logger.info("Success! Created song txt.")


def _write_headers(ctx: _Context) -> None:
    version = (
        ctx.options.txt_options.format_version
        if ctx.options and ctx.options.txt_options
        else FormatVersion.V1_0_0
    )

    if path := ctx.out.audio.path(ctx.locations, temp=True):
        _set_audio_headers(ctx, version, path)

    if path := ctx.out.video.path(ctx.locations, temp=True):
        _set_video_headers(ctx, version, path)

    if path := ctx.out.cover.path(ctx.locations, temp=True):
        _set_cover_headers(ctx, version, path)

    if path := ctx.out.background.path(ctx.locations, temp=True):
        _set_background_headers(ctx, version, path)


def _set_audio_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    match version:
        case FormatVersion.V1_0_0:
            ctx.txt.headers.mp3 = path.name
        case FormatVersion.V1_1_0:
            # write both #MP3 and #AUDIO to maximize compatibility
            ctx.txt.headers.mp3 = path.name
            ctx.txt.headers.audio = path.name
        case FormatVersion.V1_2_0:
            ctx.txt.headers.audio = path.name
            if resource := ctx.txt.meta_tags.audio or ctx.txt.meta_tags.video:
                ctx.txt.headers.audiourl = video_url_from_resource(resource)
        case _ as unreachable:
            assert_never(unreachable)


def _set_video_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    ctx.txt.headers.video = path.name
    if version == FormatVersion.V1_2_0 and (resource := ctx.txt.meta_tags.video):
        ctx.txt.headers.videourl = video_url_from_resource(resource)


def _set_cover_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    ctx.txt.headers.cover = path.name
    if (
        version == FormatVersion.V1_2_0
        and ctx.txt.meta_tags.cover
        and (url := ctx.txt.meta_tags.cover.source_url(ctx.logger))
    ):
        ctx.txt.headers.coverurl = url


def _set_background_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    ctx.txt.headers.background = path.name
    if (
        version == FormatVersion.V1_2_0
        and ctx.txt.meta_tags.background
        and (url := ctx.txt.meta_tags.background.source_url(ctx.logger))
    ):
        ctx.txt.headers.backgroundurl = url


def _maybe_write_audio_tags(ctx: _Context) -> None:
    if not (options := ctx.options.audio_options):
        return
    if not (
        audio_path_resource := ctx.out.audio.path_and_resource(ctx.locations, temp=True)
    ):
        return
    cover_path_resource = ctx.out.cover.path_and_resource(ctx.locations, temp=True)
    background_path_resource = ctx.out.background.path_and_resource(
        ctx.locations, temp=True
    )
    write_audio_tags(
        txt=ctx.txt,
        options=options,
        audio=audio_path_resource,
        cover=cover_path_resource,
        background=background_path_resource,
        logger=ctx.logger,
    )


def _maybe_write_video_tags(ctx: _Context) -> None:
    if not (options := ctx.options.video_options):
        return
    if not (
        video_path_resource := ctx.out.video.path_and_resource(ctx.locations, temp=True)
    ):
        return
    cover_path_resource = ctx.out.cover.path_and_resource(ctx.locations, temp=True)
    background_path_resource = ctx.out.background.path_and_resource(
        ctx.locations, temp=True
    )
    write_video_tags(
        txt=ctx.txt,
        options=options,
        video=video_path_resource,
        cover=cover_path_resource,
        background=background_path_resource,
        logger=ctx.logger,
    )


def _cleanup_existing_resources(ctx: _Context) -> None:
    """Delete resources that are either out of sync or will be replaced with a new one,
    and ensure kept ones are correctly named.
    """
    if not ctx.song.sync_meta:
        return
    for (old, _), out in zip(
        ctx.song.sync_meta.all_resource_files(), ctx.out, strict=False
    ):
        if not (old and (old_path := ctx.locations.current_path(file=old.fname))):
            continue
        if not out.old_fname:
            # out of sync
            if old_path.exists():
                send2trash.send2trash(old_path)
                ctx.logger.debug(f"Trashed untracked file: '{old_path}'.")
        elif out.new_fname:
            send2trash.send2trash(old_path)
            ctx.logger.debug(f"Trashed existing file: '{old_path}'.")
        else:
            target = ctx.locations.filename(ext=utils.resource_file_ending(old.fname))
            if out.old_fname != target:
                # no new file; keep existing one, but ensure correct name
                path = old_path.with_name(target)
                old_path.rename(path)
                out.old_fname = target

    return


def _persist_tempfiles(ctx: _Context) -> None:
    for temp_file in ctx.out:
        if temp_file.new_fname and (
            temp_path := temp_file.path(ctx.locations, temp=True)
        ):
            target = ctx.locations.target_path(temp_path.name)
            if target.exists():
                send2trash.send2trash(target)
                ctx.logger.debug(f"Trashed existing file: '{target}'.")
            shutil.move(temp_path, target)


def _write_sync_meta(ctx: _Context) -> None:
    old = ctx.song.sync_meta
    sync_meta_id = old.sync_meta_id if old else SyncMetaId.new()
    ctx.song.sync_meta = SyncMeta(
        sync_meta_id=sync_meta_id,
        song_id=ctx.song.song_id,
        path=ctx.locations.target_path(file=sync_meta_id.to_filename()),
        mtime=0,
        meta_tags=ctx.txt.meta_tags,
        pinned=old.pinned if old else False,
        custom_data=CustomData(old.custom_data.inner() if old else None),
    )
    ctx.song.sync_meta.txt = ctx.out.txt.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.audio = ctx.out.audio.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.video = ctx.out.video.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.cover = ctx.out.cover.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.background = ctx.out.background.to_resource_file(
        ctx.locations, temp=False
    )
    ctx.song.sync_meta.synchronize_to_file()



"""Parsed UltraStar txt file."""

from __future__ import annotations

from pathlib import Path
from typing import assert_never

import attrs

from usdb_syncer import download_options, errors
from usdb_syncer.logger import Logger
from usdb_syncer.meta_tags import MetaTags
from usdb_syncer.settings import FixLinebreaks, FixSpaces
from usdb_syncer.song_txt.headers import Headers
from usdb_syncer.song_txt.tracks import Tracks


@attrs.define()
class SongTxt:
    """A parsed .txt file of an UltraStar song."""

    headers: Headers
    notes: Tracks
    meta_tags: MetaTags
    logger: Logger

    def __str__(self) -> str:
        return f"{self.headers}\n{self.notes}"

    def unsynchronized_lyrics(self) -> str:
        track_1 = "\n".join(line.text().rstrip() for line in self.notes.track_1)
        if self.notes.track_2:
            track_2 = "\n".join(line.text().rstrip() for line in self.notes.track_2)
            return f"[{self.headers.p1}]:\n{track_1}\n\n[{self.headers.p2}]:\n{track_2}"
        return track_1

    def synchronized_lyrics(self) -> list[tuple[str, int]]:
        # format: list of tuples (lrc, millisecond)
        return [
            (
                line.text(),
                round(self.headers.bpm.beats_to_ms(line.start()) + self.headers.gap),
            )
            for line in self.notes.all_lines()
        ]

    @classmethod
    def parse(cls, value: str, logger: Logger) -> SongTxt:
        lines = [line for line in value.splitlines() if line]
        headers = Headers.parse(lines, logger)
        meta_tags = MetaTags.parse(headers.video or "", logger)
        notes = Tracks.parse(lines, logger)
        if lines:
            logger.warning(f"trailing text in song txt: '{lines}'")
        return cls(headers=headers, meta_tags=meta_tags, notes=notes, logger=logger)

    @classmethod
    def try_parse(cls, value: str, logger: Logger) -> SongTxt | None:
        try:
            return cls.parse(value, logger)
        except errors.TxtParseError:
            return None

    def maybe_split_duet_notes(self) -> None:
        if self.headers.relative and self.headers.relative.lower() == "yes":
            return
        self.notes.maybe_split_duet_notes()

    def restore_missing_headers(self) -> None:
        if self.notes.track_2:
            self.headers.p1 = self.meta_tags.player1 or "P1"
            self.headers.p2 = self.meta_tags.player2 or "P2"
        if self.meta_tags.preview is not None and self.meta_tags.preview != 0.0:
            self.headers.previewstart = self.meta_tags.preview
        if medley := self.meta_tags.medley:
            self.headers.medleystartbeat = medley.start
            self.headers.medleyendbeat = medley.end
        if self.meta_tags.tags:
            self.headers.tags = self.meta_tags.tags

    def write_to_file(self, path: Path, encoding: str, newline: str) -> None:
        with path.open(
            "w", encoding=encoding, newline=newline, errors="backslashreplace"
        ) as file:
            file.write(str(self))

    def sanitize(self, txt_options: download_options.TxtOptions | None) -> None:
        """Sanitize USDB issues and prepare for local usage."""
        self.headers.reset_file_location_headers()
        if txt_options:
            self.headers.set_version(txt_options.format_version)
            self.fix(txt_options)

    def fix(self, txt_options: download_options.TxtOptions) -> None:
        # non-optional fixes
        self.fix_relative_songs()
        self.notes.maybe_split_duet_notes()
        self.restore_missing_headers()
        self.fix_first_timestamp()
        self.fix_low_bpm()
        self.notes.fix_overlapping_and_touching_notes(self.logger)
        self.notes.fix_pitch_values(self.logger)
        self.notes.fix_apostrophes(self.logger)
        self.headers.fix_apostrophes(self.logger)
        self.notes.fix_all_caps(self.logger)
        self.headers.fix_language(self.logger)
        self.headers.fix_videogap(self.meta_tags, self.logger)
        # optional fixes
        if txt_options:
            match txt_options.fix_linebreaks:
                case FixLinebreaks.DISABLE:
                    pass
                case FixLinebreaks.USDX_STYLE:
                    self.notes.fix_linebreaks_usdx_style(self.logger)
                case FixLinebreaks.YASS_STYLE:
                    self.notes.fix_linebreaks_yass_style(self.headers.bpm, self.logger)
                case _ as unreachable:
                    assert_never(unreachable)
            if txt_options.fix_first_words_capitalization:
                self.notes.fix_first_words_capitalization(self.logger)
            if txt_options.fix_spaces != FixSpaces.DISABLE:
                self.notes.fix_spaces(txt_options.fix_spaces, self.logger)
            if txt_options.fix_quotation_marks:
                self.notes.fix_quotation_marks(
                    self.headers.main_language(), self.logger
                )

    def minimum_song_length(self) -> str:
        """Return the minimum song length based on last beat, BPM and GAP"""
        beats_secs = self.headers.bpm.beats_to_secs(self.notes.end())
        minimum_secs = round(beats_secs + self.headers.gap / 1000)
        minutes, seconds = divmod(minimum_secs, 60)

        return f"{minutes:02d}:{seconds:02d}"

    def fix_relative_songs(self) -> None:
        if not self.headers.relative:
            return

        offset = 0
        for line in self.notes.all_lines():
            for note in line.notes:
                note.start = note.start + offset

            if (line_break := line.line_break) is None:
                # must be last line, because relative timings don't support duets
                break

            line_break.previous_line_out_time = (
                line_break.previous_line_out_time + offset
            )
            if line_break.next_line_in_time is not None:
                line_break.next_line_in_time = line_break.next_line_in_time + offset
            offset = line_break.next_line_in_time or line_break.previous_line_out_time

        # remove #RELATIVE tag
        self.headers.relative = None
        self.logger.debug("FIX: Changed relative to absolute timings.")

    def fix_first_timestamp(self) -> None:
        """Shifts all notes such that the first note starts at beat zero and adjusts
        GAP accordingly
        """
        if (offset := self.notes.start()) == 0:
            # round GAP to nearest 10 ms
            self.headers.gap = int(round(self.headers.gap, -1))
            return

        for line in self.notes.all_lines():
            line.shift(-offset)

        self.headers.apply_to_medley_tags(lambda beats: beats - offset)
        offset_ms = self.headers.bpm.beats_to_ms(offset)
        self.headers.gap = int(round(self.headers.gap + offset_ms, -1))
        self.logger.debug(
            "FIX: Set first timestamp to zero and adjusted #GAP accordingly."
        )

    def fix_low_bpm(self) -> None:
        """(repeatedly) doubles BPM value and all note timings
        until the BPM is above MINIMUM_BPM"""

        if not self.headers.bpm.is_too_low():
            return

        factor = self.headers.bpm.make_large_enough()

        self.headers.apply_to_medley_tags(lambda beats: beats * factor)
        for line in self.notes.all_lines():
            line.multiply(factor)
        self.logger.debug(
            f"FIX: Increased BPM to {self.headers.bpm} (factor: {factor})"
        )
